# Generated from FullForm.g4 by ANTLR 4.7.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\33")
        buf.write("r\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\4\t\t\t\3\2\3\2\6\2\25\n\2\r\2\16\2\26\3\2\5\2\32")
        buf.write("\n\2\7\2\34\n\2\f\2\16\2\37\13\2\3\3\3\3\3\3\3\3\5\3%")
        buf.write("\n\3\3\3\3\3\3\3\3\3\3\3\7\3,\n\3\f\3\16\3/\13\3\3\4\5")
        buf.write("\4\62\n\4\3\4\3\4\5\4\66\n\4\7\48\n\4\f\4\16\4;\13\4\3")
        buf.write("\5\5\5>\n\5\3\5\3\5\3\6\5\6C\n\6\3\6\3\6\3\6\3\6\3\6\3")
        buf.write("\6\5\6K\n\6\3\7\5\7N\n\7\3\7\3\7\3\7\5\7S\n\7\3\7\5\7")
        buf.write("V\n\7\3\7\5\7Y\n\7\3\7\3\7\5\7]\n\7\3\7\5\7`\n\7\5\7b")
        buf.write("\n\7\3\b\3\b\3\b\3\b\5\bh\n\b\5\bj\n\b\3\t\3\t\5\tn\n")
        buf.write("\t\3\t\3\t\3\t\2\3\4\n\2\4\6\b\n\f\16\20\2\4\4\2\4\4\6")
        buf.write("\6\3\2\b\t\2\177\2\22\3\2\2\2\4$\3\2\2\2\6\61\3\2\2\2")
        buf.write("\b=\3\2\2\2\nJ\3\2\2\2\fa\3\2\2\2\16i\3\2\2\2\20k\3\2")
        buf.write("\2\2\22\35\5\4\3\2\23\25\7\31\2\2\24\23\3\2\2\2\25\26")
        buf.write("\3\2\2\2\26\24\3\2\2\2\26\27\3\2\2\2\27\31\3\2\2\2\30")
        buf.write("\32\5\4\3\2\31\30\3\2\2\2\31\32\3\2\2\2\32\34\3\2\2\2")
        buf.write("\33\24\3\2\2\2\34\37\3\2\2\2\35\33\3\2\2\2\35\36\3\2\2")
        buf.write("\2\36\3\3\2\2\2\37\35\3\2\2\2 !\b\3\1\2!%\5\f\7\2\"%\7")
        buf.write("\7\2\2#%\5\b\5\2$ \3\2\2\2$\"\3\2\2\2$#\3\2\2\2%-\3\2")
        buf.write("\2\2&\'\f\3\2\2\'(\7\13\2\2()\5\6\4\2)*\7\f\2\2*,\3\2")
        buf.write("\2\2+&\3\2\2\2,/\3\2\2\2-+\3\2\2\2-.\3\2\2\2.\5\3\2\2")
        buf.write("\2/-\3\2\2\2\60\62\5\4\3\2\61\60\3\2\2\2\61\62\3\2\2\2")
        buf.write("\629\3\2\2\2\63\65\7\r\2\2\64\66\5\4\3\2\65\64\3\2\2\2")
        buf.write("\65\66\3\2\2\2\668\3\2\2\2\67\63\3\2\2\28;\3\2\2\29\67")
        buf.write("\3\2\2\29:\3\2\2\2:\7\3\2\2\2;9\3\2\2\2<>\5\n\6\2=<\3")
        buf.write("\2\2\2=>\3\2\2\2>?\3\2\2\2?@\7\3\2\2@\t\3\2\2\2AC\7\21")
        buf.write("\2\2BA\3\2\2\2BC\3\2\2\2CD\3\2\2\2DE\7\3\2\2EK\7\21\2")
        buf.write("\2FG\7\3\2\2GH\7\21\2\2HI\7\3\2\2IK\7\21\2\2JB\3\2\2\2")
        buf.write("JF\3\2\2\2K\13\3\2\2\2LN\7\t\2\2ML\3\2\2\2MN\3\2\2\2N")
        buf.write("O\3\2\2\2OP\7\6\2\2PR\7\5\2\2QS\5\16\b\2RQ\3\2\2\2RS\3")
        buf.write("\2\2\2SU\3\2\2\2TV\5\20\t\2UT\3\2\2\2UV\3\2\2\2Vb\3\2")
        buf.write("\2\2WY\7\t\2\2XW\3\2\2\2XY\3\2\2\2YZ\3\2\2\2Z\\\t\2\2")
        buf.write("\2[]\5\16\b\2\\[\3\2\2\2\\]\3\2\2\2]_\3\2\2\2^`\5\20\t")
        buf.write("\2_^\3\2\2\2_`\3\2\2\2`b\3\2\2\2aM\3\2\2\2aX\3\2\2\2b")
        buf.write("\r\3\2\2\2cd\7\20\2\2dj\t\2\2\2eg\7\21\2\2fh\t\2\2\2g")
        buf.write("f\3\2\2\2gh\3\2\2\2hj\3\2\2\2ic\3\2\2\2ie\3\2\2\2j\17")
        buf.write("\3\2\2\2km\7\27\2\2ln\t\3\2\2ml\3\2\2\2mn\3\2\2\2no\3")
        buf.write("\2\2\2op\7\6\2\2p\21\3\2\2\2\27\26\31\35$-\61\659=BJM")
        buf.write("RUX\\_agim")
        return buf.getvalue()


class FullFormParser ( Parser ):

    grammarFileName = "FullForm.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'+'", "'-'", "<INVALID>", 
                     "'['", "']'", "','", "'(*'", "'*)'", "'``'", "'`'", 
                     "'''", "'\"'", "':'", "'\\'", "'^^'", "'*^'", "'.'", 
                     "'\n'", "'\uF3B1'" ]

    symbolicNames = [ "<INVALID>", "Name", "DecimalNumber", "NumberInBase", 
                      "DIGITS", "StringLiteral", "PLUS", "MINUS", "COMMENT", 
                      "LBRACKET", "RBRACKET", "COMMA", "LCOMMENT", "RCOMMENT", 
                      "DOUBLEBACKQUOTE", "BACKQUOTE", "SINGLEQUOTE", "QUOTE", 
                      "RAWCOLON", "RAWBACKSLASH", "DOUBLECARET", "ASTERISKCARET", 
                      "DOT", "NEWLINE", "CONTINUATION", "WHITESPACE" ]

    RULE_prog = 0
    RULE_expr = 1
    RULE_expressionList = 2
    RULE_symbol = 3
    RULE_context = 4
    RULE_numberLiteral = 5
    RULE_numberLiteralPrecision = 6
    RULE_numberLiteralExponent = 7

    ruleNames =  [ "prog", "expr", "expressionList", "symbol", "context", 
                   "numberLiteral", "numberLiteralPrecision", "numberLiteralExponent" ]

    EOF = Token.EOF
    Name=1
    DecimalNumber=2
    NumberInBase=3
    DIGITS=4
    StringLiteral=5
    PLUS=6
    MINUS=7
    COMMENT=8
    LBRACKET=9
    RBRACKET=10
    COMMA=11
    LCOMMENT=12
    RCOMMENT=13
    DOUBLEBACKQUOTE=14
    BACKQUOTE=15
    SINGLEQUOTE=16
    QUOTE=17
    RAWCOLON=18
    RAWBACKSLASH=19
    DOUBLECARET=20
    ASTERISKCARET=21
    DOT=22
    NEWLINE=23
    CONTINUATION=24
    WHITESPACE=25

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ProgContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(FullFormParser.ExprContext)
            else:
                return self.getTypedRuleContext(FullFormParser.ExprContext,i)


        def NEWLINE(self, i:int=None):
            if i is None:
                return self.getTokens(FullFormParser.NEWLINE)
            else:
                return self.getToken(FullFormParser.NEWLINE, i)

        def getRuleIndex(self):
            return FullFormParser.RULE_prog

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProg" ):
                listener.enterProg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProg" ):
                listener.exitProg(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitProg" ):
                return visitor.visitProg(self)
            else:
                return visitor.visitChildren(self)




    def prog(self):

        localctx = FullFormParser.ProgContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_prog)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 16
            self.expr(0)
            self.state = 27
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==FullFormParser.NEWLINE:
                self.state = 18 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 17
                        self.match(FullFormParser.NEWLINE)

                    else:
                        raise NoViableAltException(self)
                    self.state = 20 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,0,self._ctx)

                self.state = 23
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << FullFormParser.Name) | (1 << FullFormParser.DecimalNumber) | (1 << FullFormParser.DIGITS) | (1 << FullFormParser.StringLiteral) | (1 << FullFormParser.MINUS) | (1 << FullFormParser.BACKQUOTE))) != 0):
                    self.state = 22
                    self.expr(0)


                self.state = 29
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return FullFormParser.RULE_expr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class NumberContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def numberLiteral(self):
            return self.getTypedRuleContext(FullFormParser.NumberLiteralContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumber" ):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)


    class StringLiteralContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def StringLiteral(self):
            return self.getToken(FullFormParser.StringLiteral, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStringLiteral" ):
                listener.enterStringLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStringLiteral" ):
                listener.exitStringLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStringLiteral" ):
                return visitor.visitStringLiteral(self)
            else:
                return visitor.visitChildren(self)


    class SymbolLiteralContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def symbol(self):
            return self.getTypedRuleContext(FullFormParser.SymbolContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSymbolLiteral" ):
                listener.enterSymbolLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSymbolLiteral" ):
                listener.exitSymbolLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSymbolLiteral" ):
                return visitor.visitSymbolLiteral(self)
            else:
                return visitor.visitChildren(self)


    class HeadExpressionContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(FullFormParser.ExprContext,0)

        def LBRACKET(self):
            return self.getToken(FullFormParser.LBRACKET, 0)
        def expressionList(self):
            return self.getTypedRuleContext(FullFormParser.ExpressionListContext,0)

        def RBRACKET(self):
            return self.getToken(FullFormParser.RBRACKET, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterHeadExpression" ):
                listener.enterHeadExpression(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitHeadExpression" ):
                listener.exitHeadExpression(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitHeadExpression" ):
                return visitor.visitHeadExpression(self)
            else:
                return visitor.visitChildren(self)



    def expr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FullFormParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_expr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 34
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [FullFormParser.DecimalNumber, FullFormParser.DIGITS, FullFormParser.MINUS]:
                localctx = FullFormParser.NumberContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 31
                self.numberLiteral()
                pass
            elif token in [FullFormParser.StringLiteral]:
                localctx = FullFormParser.StringLiteralContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 32
                self.match(FullFormParser.StringLiteral)
                pass
            elif token in [FullFormParser.Name, FullFormParser.BACKQUOTE]:
                localctx = FullFormParser.SymbolLiteralContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 33
                self.symbol()
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 43
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,4,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FullFormParser.HeadExpressionContext(self, FullFormParser.ExprContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                    self.state = 36
                    if not self.precpred(self._ctx, 1):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                    self.state = 37
                    self.match(FullFormParser.LBRACKET)
                    self.state = 38
                    self.expressionList()
                    self.state = 39
                    self.match(FullFormParser.RBRACKET) 
                self.state = 45
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ExpressionListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return FullFormParser.RULE_expressionList

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ExpressionListedContext(ExpressionListContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ExpressionListContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(FullFormParser.ExprContext)
            else:
                return self.getTypedRuleContext(FullFormParser.ExprContext,i)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(FullFormParser.COMMA)
            else:
                return self.getToken(FullFormParser.COMMA, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExpressionListed" ):
                listener.enterExpressionListed(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExpressionListed" ):
                listener.exitExpressionListed(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpressionListed" ):
                return visitor.visitExpressionListed(self)
            else:
                return visitor.visitChildren(self)



    def expressionList(self):

        localctx = FullFormParser.ExpressionListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_expressionList)
        self._la = 0 # Token type
        try:
            localctx = FullFormParser.ExpressionListedContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 47
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << FullFormParser.Name) | (1 << FullFormParser.DecimalNumber) | (1 << FullFormParser.DIGITS) | (1 << FullFormParser.StringLiteral) | (1 << FullFormParser.MINUS) | (1 << FullFormParser.BACKQUOTE))) != 0):
                self.state = 46
                self.expr(0)


            self.state = 55
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==FullFormParser.COMMA:
                self.state = 49
                self.match(FullFormParser.COMMA)
                self.state = 51
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << FullFormParser.Name) | (1 << FullFormParser.DecimalNumber) | (1 << FullFormParser.DIGITS) | (1 << FullFormParser.StringLiteral) | (1 << FullFormParser.MINUS) | (1 << FullFormParser.BACKQUOTE))) != 0):
                    self.state = 50
                    self.expr(0)


                self.state = 57
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SymbolContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return FullFormParser.RULE_symbol

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ContextNameContext(SymbolContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.SymbolContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Name(self):
            return self.getToken(FullFormParser.Name, 0)
        def context(self):
            return self.getTypedRuleContext(FullFormParser.ContextContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterContextName" ):
                listener.enterContextName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitContextName" ):
                listener.exitContextName(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitContextName" ):
                return visitor.visitContextName(self)
            else:
                return visitor.visitChildren(self)



    def symbol(self):

        localctx = FullFormParser.SymbolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_symbol)
        try:
            localctx = FullFormParser.ContextNameContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 59
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.state = 58
                self.context()


            self.state = 61
            self.match(FullFormParser.Name)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ContextContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return FullFormParser.RULE_context

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class SimpleContextContext(ContextContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ContextContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Name(self):
            return self.getToken(FullFormParser.Name, 0)
        def BACKQUOTE(self, i:int=None):
            if i is None:
                return self.getTokens(FullFormParser.BACKQUOTE)
            else:
                return self.getToken(FullFormParser.BACKQUOTE, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSimpleContext" ):
                listener.enterSimpleContext(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSimpleContext" ):
                listener.exitSimpleContext(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSimpleContext" ):
                return visitor.visitSimpleContext(self)
            else:
                return visitor.visitChildren(self)


    class CompoundContextContext(ContextContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.ContextContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def Name(self, i:int=None):
            if i is None:
                return self.getTokens(FullFormParser.Name)
            else:
                return self.getToken(FullFormParser.Name, i)
        def BACKQUOTE(self, i:int=None):
            if i is None:
                return self.getTokens(FullFormParser.BACKQUOTE)
            else:
                return self.getToken(FullFormParser.BACKQUOTE, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCompoundContext" ):
                listener.enterCompoundContext(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCompoundContext" ):
                listener.exitCompoundContext(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitCompoundContext" ):
                return visitor.visitCompoundContext(self)
            else:
                return visitor.visitChildren(self)



    def context(self):

        localctx = FullFormParser.ContextContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_context)
        self._la = 0 # Token type
        try:
            self.state = 72
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
            if la_ == 1:
                localctx = FullFormParser.SimpleContextContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 64
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==FullFormParser.BACKQUOTE:
                    self.state = 63
                    self.match(FullFormParser.BACKQUOTE)


                self.state = 66
                self.match(FullFormParser.Name)
                self.state = 67
                self.match(FullFormParser.BACKQUOTE)
                pass

            elif la_ == 2:
                localctx = FullFormParser.CompoundContextContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 68
                self.match(FullFormParser.Name)
                self.state = 69
                self.match(FullFormParser.BACKQUOTE)
                self.state = 70
                self.match(FullFormParser.Name)
                self.state = 71
                self.match(FullFormParser.BACKQUOTE)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberLiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return FullFormParser.RULE_numberLiteral

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class NumberBaseNContext(NumberLiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.NumberLiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def DIGITS(self):
            return self.getToken(FullFormParser.DIGITS, 0)
        def NumberInBase(self):
            return self.getToken(FullFormParser.NumberInBase, 0)
        def MINUS(self):
            return self.getToken(FullFormParser.MINUS, 0)
        def numberLiteralPrecision(self):
            return self.getTypedRuleContext(FullFormParser.NumberLiteralPrecisionContext,0)

        def numberLiteralExponent(self):
            return self.getTypedRuleContext(FullFormParser.NumberLiteralExponentContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumberBaseN" ):
                listener.enterNumberBaseN(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumberBaseN" ):
                listener.exitNumberBaseN(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumberBaseN" ):
                return visitor.visitNumberBaseN(self)
            else:
                return visitor.visitChildren(self)


    class NumberBaseTenContext(NumberLiteralContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a FullFormParser.NumberLiteralContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def DIGITS(self):
            return self.getToken(FullFormParser.DIGITS, 0)
        def DecimalNumber(self):
            return self.getToken(FullFormParser.DecimalNumber, 0)
        def MINUS(self):
            return self.getToken(FullFormParser.MINUS, 0)
        def numberLiteralPrecision(self):
            return self.getTypedRuleContext(FullFormParser.NumberLiteralPrecisionContext,0)

        def numberLiteralExponent(self):
            return self.getTypedRuleContext(FullFormParser.NumberLiteralExponentContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumberBaseTen" ):
                listener.enterNumberBaseTen(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumberBaseTen" ):
                listener.exitNumberBaseTen(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumberBaseTen" ):
                return visitor.visitNumberBaseTen(self)
            else:
                return visitor.visitChildren(self)



    def numberLiteral(self):

        localctx = FullFormParser.NumberLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_numberLiteral)
        self._la = 0 # Token type
        try:
            self.state = 95
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,17,self._ctx)
            if la_ == 1:
                localctx = FullFormParser.NumberBaseNContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 75
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==FullFormParser.MINUS:
                    self.state = 74
                    self.match(FullFormParser.MINUS)


                self.state = 77
                self.match(FullFormParser.DIGITS)
                self.state = 78
                self.match(FullFormParser.NumberInBase)
                self.state = 80
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,12,self._ctx)
                if la_ == 1:
                    self.state = 79
                    self.numberLiteralPrecision()


                self.state = 83
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,13,self._ctx)
                if la_ == 1:
                    self.state = 82
                    self.numberLiteralExponent()


                pass

            elif la_ == 2:
                localctx = FullFormParser.NumberBaseTenContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 86
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==FullFormParser.MINUS:
                    self.state = 85
                    self.match(FullFormParser.MINUS)


                self.state = 88
                _la = self._input.LA(1)
                if not(_la==FullFormParser.DecimalNumber or _la==FullFormParser.DIGITS):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 90
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,15,self._ctx)
                if la_ == 1:
                    self.state = 89
                    self.numberLiteralPrecision()


                self.state = 93
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,16,self._ctx)
                if la_ == 1:
                    self.state = 92
                    self.numberLiteralExponent()


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberLiteralPrecisionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOUBLEBACKQUOTE(self):
            return self.getToken(FullFormParser.DOUBLEBACKQUOTE, 0)

        def DecimalNumber(self):
            return self.getToken(FullFormParser.DecimalNumber, 0)

        def DIGITS(self):
            return self.getToken(FullFormParser.DIGITS, 0)

        def BACKQUOTE(self):
            return self.getToken(FullFormParser.BACKQUOTE, 0)

        def getRuleIndex(self):
            return FullFormParser.RULE_numberLiteralPrecision

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumberLiteralPrecision" ):
                listener.enterNumberLiteralPrecision(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumberLiteralPrecision" ):
                listener.exitNumberLiteralPrecision(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumberLiteralPrecision" ):
                return visitor.visitNumberLiteralPrecision(self)
            else:
                return visitor.visitChildren(self)




    def numberLiteralPrecision(self):

        localctx = FullFormParser.NumberLiteralPrecisionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_numberLiteralPrecision)
        self._la = 0 # Token type
        try:
            self.state = 103
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [FullFormParser.DOUBLEBACKQUOTE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 97
                self.match(FullFormParser.DOUBLEBACKQUOTE)
                self.state = 98
                _la = self._input.LA(1)
                if not(_la==FullFormParser.DecimalNumber or _la==FullFormParser.DIGITS):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass
            elif token in [FullFormParser.BACKQUOTE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 99
                self.match(FullFormParser.BACKQUOTE)
                self.state = 101
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
                if la_ == 1:
                    self.state = 100
                    _la = self._input.LA(1)
                    if not(_la==FullFormParser.DecimalNumber or _la==FullFormParser.DIGITS):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberLiteralExponentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ASTERISKCARET(self):
            return self.getToken(FullFormParser.ASTERISKCARET, 0)

        def DIGITS(self):
            return self.getToken(FullFormParser.DIGITS, 0)

        def PLUS(self):
            return self.getToken(FullFormParser.PLUS, 0)

        def MINUS(self):
            return self.getToken(FullFormParser.MINUS, 0)

        def getRuleIndex(self):
            return FullFormParser.RULE_numberLiteralExponent

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumberLiteralExponent" ):
                listener.enterNumberLiteralExponent(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumberLiteralExponent" ):
                listener.exitNumberLiteralExponent(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumberLiteralExponent" ):
                return visitor.visitNumberLiteralExponent(self)
            else:
                return visitor.visitChildren(self)




    def numberLiteralExponent(self):

        localctx = FullFormParser.NumberLiteralExponentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_numberLiteralExponent)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 105
            self.match(FullFormParser.ASTERISKCARET)
            self.state = 107
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==FullFormParser.PLUS or _la==FullFormParser.MINUS:
                self.state = 106
                _la = self._input.LA(1)
                if not(_la==FullFormParser.PLUS or _la==FullFormParser.MINUS):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 109
            self.match(FullFormParser.DIGITS)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.expr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expr_sempred(self, localctx:ExprContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 1)
         







"""
Classes and utilities for handling built-in symbols which do not have an
associated `SymbolNode` representing its declaration. We call such a symbol
*virtual* because its declaration does not actually exist in the source,
because built-in functions and symbols are never explicitly declared.

Currently this is just a wrapper that provides access to the services of
WolframLanguageData.
"""

from FoxySheep.scoping import Symbol
import FoxySheep.utils.WolframLanguageData as wld


def find_virtual_symbol(cls, name: str) -> Symbol:
    """
    Returns a virtual symbol representing a built-in symbol or function.

    :param name: String, the name of the symbol.
    :return: A VirtualNode
    """

    return wld.find_symbol(name)


def get_operator_name(symbolic_rep: str) -> str:
    """
    Given the symbolic string representation for a built-in
    operator, returns its canonical name.

    For example:

        >>> get_operator_name('/@')
        'Map'

    :param symbolic_rep: The symbolic string representation of the operator.
    :return: The canonical name of the operator as a string.
    """

    return wld.get_operator_name()



from FoxySheep.tree import treeNode, SymbolNode


class BlankAbstractNode(treeNode):
    """
    `BlankAbstractNode` is the superclass of Blank*Node, each of which has
    one optional child called `required_head`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add_named_children(['required_head'])



from enum import IntEnum, auto as _auto
import FoxySheep.tree
from FoxySheep.tree import treeNode, FunctionDeclarationNode

class SymbolAttributes(IntEnum):
    Constant        = _auto()
    Flat            = _auto()
    HoldAll         = _auto()
    HoldAllComplete = _auto()
    HoldFirst       = _auto()
    HoldRest        = _auto()
    Listable        = _auto()
    Locked          = _auto()
    NHoldAll        = _auto()
    NHoldFirst      = _auto()
    NHoldRest       = _auto()
    NumericFunction = _auto()
    OneIdentity     = _auto()
    Orderless       = _auto()
    Protected       = _auto()
    ReadProtected   = _auto()
    SequenceHold    = _auto()
    Stub            = _auto()
    Temporary       = _auto()


class Symbol(FoxySheep.tree.Symbol):
    """
    Holds information about a particular symbol in a particular scope. In
    particular, it manages

        * OwnValues: Values without square brackets []. Only the first
        OwnValue is accessible via evaluating the symbol. Other OwnValues can
        only be set or accessed by manipulating OwnValues[name] itself.

        * DownValues: Function definitions, i.e., anything accessed with [].
        A DownValue can be considered a function declaration; it is the
        treeNode for the assignment that "creates" the DownValue.

        * DefaultValues: Default values for optional arguments.

        * Options: Option defaults for the symbol.

        * Attributes: A set of `SymbolAttributes`.

        * Messages: Messages of the form `symbol::tag`.
    """

    def __init__(self,
                 identifier: str,  # Required
                 own_values: list = None,
                 down_values: list = None,
                 default_values: dict = None,
                 # Values
                 options: set = None,
                 attributes: set = None,
                 messages: set = None,
                 # Conditions
                 required_head: 'Symbol' = None
                 # Not implemented:
                 # up_values: list = None,
                 # sub_values: list = None,
                 # n_values: list = None,
                 # format_values: list = None,
                 ):
        """


        :param identifier:
        :param own_values:
        :param down_values:
        :param default_values:
        :param options:
        :param attributes:
        :param messages:
        :param required_head:
        """

        self._identifier = identifier

        # We lazily create these container classes.
        self._own_values        = own_values
        self._down_values       = down_values
        self._default_values 	= default_values
        self._options 	        = options
        self._attributes 	    = attributes
        self._messages 	        = messages
        self._required_head     = required_head

        # Not implemented.
        # TODO: Implement UpValues.
        # self._up_values         = up_values
        # TODO: Implement SubValues.
        # self._sub_values        = sub_values
        # TODO: Implement NValues.
        # self._n_values          = n_values
        # TODO: Implement FormatValues.
        # self._format_values     = format_values

    @property
    def default(self):
        # Default values are associated to the position of a parameter. It is
        # possible to have a single default value without a position
        # designation. Internally, we associate the default value with
        # position -1.
        return self.get_default_at(-1)

    @default.setter
    def default(self, value):
        self.set_default_at(-1, value)

    def get_default_at(self, position: int):
        # Default values are associated to the position of a parameter. The
        # positions with defaults are sparse, so we use a dict rather than a
        # list to represent them.
        try:
            return self._default_values[position]
        except KeyError:
            pass
        return None

    def set_default_at(self, position: int, value: treeNode):
        if not self._default_values:
            self._default_values = dict()
        self._default_values[position] = value

    @property
    def identifier(self):
        return self._identifier

    @property
    def own_value(self):
        """
        Fetch the OwnValue of this symbol.

        TODO: Support multiple OwnValues.

        :return: An OwnValue, if one exists; otherwise None.
        """
        if self._own_values:
            return self._own_values[0]
        return None

    def find_matching_down_value(self, usage: treeNode):
        """
        Given a usage `f[3, 5]`, find the first DownValue of f matching the
        usage.

        :param usage: A function call.
        :return:
        """

        # TODO: Implement using Utils.PatternMatcher.match_pattern().

        return self._default_values[-1]

    def add_down_value(self, value: FunctionDeclarationNode):
        """
        Note: A DownValue is a function declaration.

        Adds a DownValue to the list of DownValues associated to this symbol.
        A DownValue must be a subclass of FunctionDeclarationNode. As with
        Mathematica, new DownValues overwrite existing DownValues having the
        same LHS, and DownValues are ordered according to the order they are
        added, except that DownValues with more specific LHS are ordered
        before DownValues with more general LHS.

        :param value: The DownValue to be added.
        """

        # TODO: Implement correct ordering of DownValues using Utils.PatternMatcher.match_pattern().

        self._default_values.append(value)



from antlr4.ParserRuleContext import ParserRuleContext
from antlr4 import TerminalNode, ParseTreeWalker, ParseTreeListener
import sys

INDENTATION_SPACE = "  "
class PrettyPrinter(ParseTreeListener):
    def __init__(self, rule_names):
        self.rule_names = rule_names
        self.builder = ""
        self.indentation = 0
        self.kid_number = 0

    def visitTerminal(self, node: TerminalNode) -> None:
        if self.builder:
            if self.indentation:
                self.builder += "\n"

            self.builder += (INDENTATION_SPACE * self.indentation)

        s = str(node)
        if not s.isidentifier():
            s = repr(str(node))
        if self.kid_number:
            self.builder += f"{self.kid_number}. "
        self.kid_number += 1
        self.builder += s + "\n"
        # builder.append(Utils.escapeWhitespace(Trees.getNodeText(node, ruleNames), false));

    def visitErrorNode(self, node) -> None:
        if self.builder:
            if self.indentation:
                self.builder += "\n"
            self.builder += (INDENTATION_SPACE * self.indentation)

        self.builder += f"{self.kid_number}. "
        self.kid_number += 1
        self.builder += str(node)
        # builder.append(Utils.escapeWhitespace(Trees.getNodeText(node, ruleNames), false));

    def enterEveryRule(self, ctx: ParserRuleContext) -> None:
        if self.indentation:
            self.builder += "\n"

        self.builder += (INDENTATION_SPACE * self.indentation)
        self.indentation += 1

        child_count = ctx.getChildCount()
        if child_count and ctx.parentCtx and ctx.parentCtx.getChildCount() > 1:
            self.builder += f"{self.kid_number}. "

        rule_index = ctx.getRuleIndex()
        if rule_index is not None and rule_index < len(self.rule_names):
            rule_name = self.rule_names[rule_index]
        else:
            rule_name = str(rule_index)

        nt_name = ctx.__class__.__name__
        if nt_name.endswith("Context"):
            nt_name = nt_name[:-len("Context")]
        name = "<%s:%s> " % (rule_name, nt_name)
        self.kid_number += 1
        self.builder += name
        if child_count:
            self.builder += f"[{child_count}]"
            self.kid_number = 0

    def exitEveryRule(self, ctx: ParserRuleContext) -> None:
        if ctx.getChildCount():
            self.indentation -= 1

class PrettyPrinterCompact(ParseTreeListener):
    def __init__(self, rule_names):
        self.rule_names = rule_names
        self.builder = ""

    def visitTerminal(self, node: TerminalNode) -> None:
        if self.builder:
            self.builder += " "
        s = str(node)
        if not s.isidentifier():
            s = repr(str(node))
        self.builder += s
        # builder.append(Utils.escapeWhitespace(Trees.getNodeText(node, ruleNames), false));

    def visitErrorNode(self, node) -> None:
        if self.builder:
            self.builder += " "

        self.builder += f"{self.kid_number}. "
        self.kid_number += 1
        self.builder += str(node)
        # builder.append(Utils.escapeWhitespace(Trees.getNodeText(node, ruleNames), false));

    def enterEveryRule(self, ctx: ParserRuleContext) -> None:
        if self.builder:
            self.builder += " "

        child_count = ctx.getChildCount()
        if child_count:
            self.builder += "("

        rule_index = ctx.getRuleIndex()
        if rule_index is not None and rule_index < len(self.rule_names):
            rule_name = self.rule_names[rule_index]
        else:
            rule_name = str(rule_index)

        nt_name = ctx.__class__.__name__
        if nt_name.endswith("Context"):
            nt_name = nt_name[:-len("Context")]
        name = "<%s:%s> " % (rule_name, nt_name)

        self.builder += name

    def exitEveryRule(self, ctx: ParserRuleContext) -> None:
        if ctx.getChildCount():
            self.builder += ")"

def pretty_print_string(tree, rule_names, compact=False) -> str:
    walker = ParseTreeWalker()
    pp_fn = PrettyPrinterCompact if compact else PrettyPrinter
    pretty_printer = pp_fn(rule_names)
    walker.walk(pretty_printer, tree)
    return pretty_printer.builder

def pretty_print_string_compact(*args) -> None:
    pretty_print_string(*args, compact=True)

def pretty_print(*args, **kwargs) -> None:
    out = kwargs.get("out", sys.stdout)
    out.write(pretty_print_string(*args, **kwargs) + "\n")

def pretty_print_compact(*args, **kwargs) -> None:
    kwargs["compact"] = True
    pretty_print(*args, **kwargs)



"""
A command line program to perform Mathematica translations.

When installed run `foxy-sheep --help` for options
"""
import click
import importlib
import sys
import traceback
from typing import Any, Callable, Optional
from FoxySheep.parser import ff_parse_tree_from_string, parse_tree_from_string, parse_tree_from_string_pp
from FoxySheep.emitter.python import input_form_to_python
from FoxySheep.emitter.full_form import input_form_to_full_form
from FoxySheep.transform.if_transform import input_form_post
from FoxySheep.tree.pretty_printer import pretty_print
from FoxySheep.version import VERSION as __version__


# TODO: we could put this in a class and then one could have many REPLs.
out_results = []

eval_namespace = {
    "out_results": out_results,
    "missing_modules": [],
}

def setup_session():
    for importname in ("decimal","math", "matplotlib.pyplot"):
        try:
            eval_namespace[importname] = importlib.import_module(importname)
        except ImportError:
            print(f"Error importing {importname}; translations using this module will fail.")
            eval_namespace["missing_modules"].append(importname)

def Out(i: Optional[int]=None) -> Any:
    if i is None:
        i = -1
    if not len(out_results):
        raise RuntimeError("No prior input")
    return out_results[i]

def REPL(parse_tree_fn: Callable, output_style_fn, session, show_tree_fn=None) -> None:
    """
    Read Eval Print Loop (REPL) for Mathematica translations
    """
    print(
        "Enter a Mathematica expression. Enter either an empty line, Ctrl-C, or Ctrl-D to exit."
    )
    while True:
        try:
            user_in = input("in:= ")
        except (KeyboardInterrupt, EOFError):
            break
        if user_in == "":
            break

        try:
            results = output_style_fn(user_in, parse_tree_fn, show_tree_fn)
        except:
            traceback.print_exc(5)
            continue

        print(results)
        if session:
            try:
                x = eval(results, None, eval_namespace)
            except:
                print(sys.exc_info()[1])
            else:
                print(f"Out[{len(out_results)}]={x}")
                out_results.append(x)
                # from pprint import pprint
                # pprint(out_results)
                pass
            pass
        pass


@click.command()
@click.option(
    "--repl/--no-repl", default=True, required=False, type=bool, help="Go into REPL",
)
@click.option(
    "-t", "--tree",
    type=click.Choice(["full", "compact"], case_sensitive=False),
    required=False,
    help="show parse tree(s) created",
)
@click.option(
    "-i",
    "--input-style",
    type=click.Choice(["InputForm", "FullForm"], case_sensitive=False),
    required=False,
)
@click.option(
    "-o",
    "--output-style",
    type=click.Choice(["Python", "FullForm"], case_sensitive=False),
    required=False,
)
@click.option(
    "-s",
    "--session/--no-session",
    default=None,
    required=False,
    help="In REPL, evaluate the translation and in REPL session"
)
@click.option("-e", "--expr", help="translate *expr*", required=False)
@click.version_option(version=__version__)
def main(repl: bool, tree, input_style, output_style, session: bool, expr: str):
    parse_tree_fn = (
        ff_parse_tree_from_string
        if input_style and input_style.lower() == "fullform"
        else parse_tree_from_string_pp
    )

    if tree == "full":
        show_tree_fn = lambda tree, rule_names: pretty_print(tree, rule_names, compact=False)
    elif tree == "compact":
        show_tree_fn = lambda tree, rule_names: pretty_print(tree, rule_names, compact=True)
    else:
        show_tree_fn = None

    output_style_fn = input_form_to_full_form
    if output_style and output_style.lower() == "python":
        output_style_fn = input_form_to_python
        parse_tree_fn = parse_tree_from_string
        if session == None and not expr:
            session = True
            pass
        pass
    elif session:
        print("--session option is only valid for Python output. Option ignored.")
        session = False

    if expr:
        if session:
            print("--session option is only valid in a REPL. Option ignored.")
        print(output_style_fn(expr, parse_tree_fn, show_tree_fn))
    elif repl:
        setup_session()
        REPL(parse_tree_fn, output_style_fn, session, show_tree_fn)
    else:
        print("Something went wrong")


if __name__ == "__main__":
    main()



"""Robert Jacobson's ANTL4-based translator parser for Mathematica."""

from FoxySheep.generated.InputFormLexer import InputFormLexer
from FoxySheep.generated.InputFormParser import InputFormParser
from FoxySheep.generated.InputFormListener import InputFormListener
from FoxySheep.generated.InputFormVisitor import InputFormVisitor

from FoxySheep.generated.FullFormLexer import FullFormLexer
from FoxySheep.generated.FullFormParser import FullFormParser
from FoxySheep.generated.FullFormListener import FullFormListener
from FoxySheep.generated.FullFormVisitor import FullFormVisitor

from FoxySheep.emitter.full_form import FullFormEmitter, input_form_to_full_form
from FoxySheep.emitter.python import input_form_to_python

from FoxySheep.parser import (
    ff_parse_tree_from_string,
    if2ff,
    parse_tree_from_string,
)
from FoxySheep.post_parser import PostParser
from FoxySheep.tree.pretty_printer import (
    pretty_print,
    pretty_print_string,
    PrettyPrinter,
    PrettyPrinterCompact,
)
from FoxySheep.tree import SymbolNode
from FoxySheep.__main__ import REPL

from FoxySheep.version import VERSION

# This ensures VERSION will appear in pydoc
__version__ = VERSION



from AST import ASTNode


class OptionalNode(ASTNode):
    """
    `Optional[pattern_, default_.]`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add_named_children(['pattern', 'default'])



from antlr4 import ParseTreeWalker
from FoxySheep.post_parser import PostParser

def input_form_post(tree):
    """Post process the parse tree. In particular, flatten some flat
    operators. Some operators appear in the source text without any
    explicit associativity, such as `Plus`, and are parsed into arbitrary
    tree structures.
    """

    walker = ParseTreeWalker()
    post_parser = PostParser()
    walker.walk(post_parser, tree)

    # The PostParser can restructure the tree in a way that changes the root.
    if tree.parentCtx is not None:
        tree = tree.parentCtx
    return tree



# Generated from FullForm.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .FullFormParser import FullFormParser
else:
    from FullFormParser import FullFormParser

# This class defines a complete listener for a parse tree produced by FullFormParser.
class FullFormListener(ParseTreeListener):

    # Enter a parse tree produced by FullFormParser#prog.
    def enterProg(self, ctx:FullFormParser.ProgContext):
        pass

    # Exit a parse tree produced by FullFormParser#prog.
    def exitProg(self, ctx:FullFormParser.ProgContext):
        pass


    # Enter a parse tree produced by FullFormParser#Number.
    def enterNumber(self, ctx:FullFormParser.NumberContext):
        pass

    # Exit a parse tree produced by FullFormParser#Number.
    def exitNumber(self, ctx:FullFormParser.NumberContext):
        pass


    # Enter a parse tree produced by FullFormParser#StringLiteral.
    def enterStringLiteral(self, ctx:FullFormParser.StringLiteralContext):
        pass

    # Exit a parse tree produced by FullFormParser#StringLiteral.
    def exitStringLiteral(self, ctx:FullFormParser.StringLiteralContext):
        pass


    # Enter a parse tree produced by FullFormParser#SymbolLiteral.
    def enterSymbolLiteral(self, ctx:FullFormParser.SymbolLiteralContext):
        pass

    # Exit a parse tree produced by FullFormParser#SymbolLiteral.
    def exitSymbolLiteral(self, ctx:FullFormParser.SymbolLiteralContext):
        pass


    # Enter a parse tree produced by FullFormParser#HeadExpression.
    def enterHeadExpression(self, ctx:FullFormParser.HeadExpressionContext):
        pass

    # Exit a parse tree produced by FullFormParser#HeadExpression.
    def exitHeadExpression(self, ctx:FullFormParser.HeadExpressionContext):
        pass


    # Enter a parse tree produced by FullFormParser#ExpressionListed.
    def enterExpressionListed(self, ctx:FullFormParser.ExpressionListedContext):
        pass

    # Exit a parse tree produced by FullFormParser#ExpressionListed.
    def exitExpressionListed(self, ctx:FullFormParser.ExpressionListedContext):
        pass


    # Enter a parse tree produced by FullFormParser#ContextName.
    def enterContextName(self, ctx:FullFormParser.ContextNameContext):
        pass

    # Exit a parse tree produced by FullFormParser#ContextName.
    def exitContextName(self, ctx:FullFormParser.ContextNameContext):
        pass


    # Enter a parse tree produced by FullFormParser#SimpleContext.
    def enterSimpleContext(self, ctx:FullFormParser.SimpleContextContext):
        pass

    # Exit a parse tree produced by FullFormParser#SimpleContext.
    def exitSimpleContext(self, ctx:FullFormParser.SimpleContextContext):
        pass


    # Enter a parse tree produced by FullFormParser#CompoundContext.
    def enterCompoundContext(self, ctx:FullFormParser.CompoundContextContext):
        pass

    # Exit a parse tree produced by FullFormParser#CompoundContext.
    def exitCompoundContext(self, ctx:FullFormParser.CompoundContextContext):
        pass


    # Enter a parse tree produced by FullFormParser#NumberBaseN.
    def enterNumberBaseN(self, ctx:FullFormParser.NumberBaseNContext):
        pass

    # Exit a parse tree produced by FullFormParser#NumberBaseN.
    def exitNumberBaseN(self, ctx:FullFormParser.NumberBaseNContext):
        pass


    # Enter a parse tree produced by FullFormParser#NumberBaseTen.
    def enterNumberBaseTen(self, ctx:FullFormParser.NumberBaseTenContext):
        pass

    # Exit a parse tree produced by FullFormParser#NumberBaseTen.
    def exitNumberBaseTen(self, ctx:FullFormParser.NumberBaseTenContext):
        pass


    # Enter a parse tree produced by FullFormParser#numberLiteralPrecision.
    def enterNumberLiteralPrecision(self, ctx:FullFormParser.NumberLiteralPrecisionContext):
        pass

    # Exit a parse tree produced by FullFormParser#numberLiteralPrecision.
    def exitNumberLiteralPrecision(self, ctx:FullFormParser.NumberLiteralPrecisionContext):
        pass


    # Enter a parse tree produced by FullFormParser#numberLiteralExponent.
    def enterNumberLiteralExponent(self, ctx:FullFormParser.NumberLiteralExponentContext):
        pass

    # Exit a parse tree produced by FullFormParser#numberLiteralExponent.
    def exitNumberLiteralExponent(self, ctx:FullFormParser.NumberLiteralExponentContext):
        pass





# 'Node' is appended to every AST class name. There should be a class for
# every Wolfram Language construct we wish to transform or otherwise interact
# with.

from enum import IntEnum, auto

from AST import ASTNode


class Props(IntEnum):
    NamedChildren = auto()
    SuperClass = auto()
    DoNotCreate = auto()


ast_classes = \
{
    'pattern/BlankAbstract': {Props.NamedChildren: ['required_head']},
    'pattern/Blank': {Props.SuperClasses: ['BlankAbstractNode']},
    'pattern/BlankSequence': {Props.SuperClasses: ['BlankAbstractNode']},
    'pattern/BlankNullSequence': {Props.SuperClasses: ['BlankAbstractNode']},
    'pattern/Pattern': {Props.NamedChildren: ['identifier', 'pattern']},
    'Symbol': {Props.DoNotCreate: None},

}



"""
Classes for Scoping.
"""
from typing import List

import FoxySheep.tree

from enum import Enum, auto as _auto
from FoxySheep.scoping import Symbol


class ScopeType(Enum):
    Empty = _auto()
    Table = _auto()
    Solve = _auto()
    Integrate = _auto()
    Limit = _auto()
    Plot = _auto()
    Manipulate = _auto()
    Module = _auto()
    Function = _auto()
    AnonymousFunction = _auto()
    Rule = _auto()


class Scope(FoxySheep.tree.Scope):
    """
    Represents the local scope of a Scoping construct, e.g. SetDelayed, Table

    Properties:
        `parent`: The Scoping construct to which this scope belongs; a
        DownValue.

        `parent_scope`: A reference to the containing scope.

        `self.symbols = dict()`: Holds the local variables.

    The next four properties derive from the scope type as given by the
    Wolfram Language built-in `_SyntaxInformation[symbol]`:

        `self.scope_type = scope_type`:  A value from the ScopeType enum
        describing how the local variables are declared.

        `self.scope_start = scope_start`: The starting position of the local
        variable list.

        `self.scope_end = scope_end` The end position of the local variable
        list.

        `self.body_position = body_position` The position of the "body"
        expression existing within this scope.
    """

    def __init__(
        self,
        parent: FoxySheep.tree.treeNode = None,
        parent_scope: "Scope" = None,
        scope_type: int = None,
        scope_start: int = -1,
        scope_end: int = -1,
        body_location: int = -1,
    ):
        """
        Create a new local scope.

        :param parent: The Scoping construct to which this scope belongs; a
        DownValue.
        :param parent_scope: A reference to the containing scope, or None if
        this is the global scope.
        :param scope_type: A value from the ScopeType enum.
        :param scope_start: The starting position of the local variable list.
        :param scope_end: The end position of the local variable list.
        :param body_location: The position of the "body" expression existing
        within this scope.
        """

        self.parent = parent
        self.parent_scope = parent_scope
        self._symbols = dict()
        self.scope_type = scope_type
        self.scope_start = scope_start
        self.scope_end = scope_end
        self.body_position = body_location

    def find_symbol(self, identifier: str) -> Symbol:
        return self._symbols.get(identifier, default=None)

    def add_symbol(self, symbol: Symbol):
        """
        Adds a new symbol to the Scope.

        A symbol added to a local scope must be declared by virtue of a
        Scoping construct (`Module`, `Table`, `SetDelayed`, etc.), as all
        other symbols are added to the `GlobalScope`.

        :param symbol: The Symbol object to add to the scope.
        :return:
        """

        # TODO: Check that another symbol with the same identifier is not already in this scope.
        self._symbols[symbol.identifier] = symbol

    def add_symbols(self, symbols: List[Symbol]):
        for symbol in symbols:
            self.add_symbol(symbol)


class GlobalScope(Scope):
    """This class stores user-declared symbols in the global scope. Note that
    virtual symbols are not stored in this class but rather are managed by a
    VirtualSymbolProvider.

    Note that any new symbol that is not created by a Scoping construct (
    `Module`, `Table`, etc.) is created in the `GlobalScope`.
    """

    def __init__(self):
        super().__init__(parent=None)
        # rocky: VirtualSymbolProvider is missing in FoxySheep
        # self._virtual_symbol_provider = VirtualSymbolProvider()

    def find_symbol(self, identifier: str) -> Symbol:
        """
        First searches for `identifier` in the list of user-defined symbols
        in the global namespace. If no concrete declaration for the symbol is
        found, we look in the database of built-in symbols, returning a
        VirtualNode if it's found. Otherwise, we return None.

        :param identifier: String, the name of the symbol we are searching
        for.
        :return: A concrete symbol declaration, or a VirtualNode
        representing the "declaration" of a built-in symbol, or None if the
        symbol isn't found.
        """

        # Search the global namespace for user-declared symbols.
        found = self._symbols.get(identifier, default=None)
        if found:
            return found

        # Now ask our VirtualSymbolProvider to search the database of
        # built-in symbols.
        return self._virtual_symbol_provider.find_virtual_symbol(identifier)


class ScopeStack(object):
    """
    A stack data structure that keeps track of the (linear) nested scopes as
    we walk the AST.

    Properties:
        `stack` A Python list of scopes.
    """

    def __init__(self, global_scope: GlobalScope = None):
        self._stack: List[Scope] = []
        if global_scope:
            self._stack.append(global_scope)
        else:
            self._stack.append(GlobalScope())

    def push(self, scope: Scope):
        self._stack.append(scope)

    def pop(self) -> Scope:
        return self._stack.pop()

    def find_symbol(self, identifier: str) -> Symbol:

        for scope in reversed(self._stack):
            found = scope.find_symbol(identifier)
            if found:
                return found

        # The symbol isn't found in any local scope.
        return None



from AST import ASTNode
from FoxySheep.scoping.Symbol import Symbol


class FunctionSignatureNode(ASTNode):
    """
    An abstraction representing a function signature, including the symbol
    for the function name, the parameters and their types, and default values.
    """

    def __init__(self, symbol: Symbol = None, **kwargs):
        """

        :param symbol: A `Symbol` for the head (name) of the function.
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.symbol = symbol



# This class is a subclass FoxySheepListener that rewrites the
# parse tree generated by ANTLR4. It "flattens" flat operators that ANTLR
# parses as left associative.

from antlr4.ParserRuleContext import ParserRuleContext
from antlr4.tree.Tree import TerminalNodeImpl, ParseTree
from antlr4.Token import CommonToken
from FoxySheep.generated.InputFormListener import InputFormListener
from FoxySheep.generated.InputFormParser import InputFormParser


def addChild(parent:ParserRuleContext, child:ParseTree, i:int = None):
    """Does what RuleContext.addChild is supposed to do. The ANTLR4 Python3 target
    doesn't follow the same API as the Java target. Some classes in the hierarchy
    don't have addChild, while others do."""
    try:
        if i is None:
            parent.children.append(child)
        else:
            parent.children.insert(i, child)
    except:
        parent.children = [child]


def adopt(parent:ParserRuleContext, child:ParseTree):
    """Convenience function to create a bidirectional parent-child relationship."""
    addChild(parent, child)
    child.parentCtx = parent


def makeNumber(parent:ParserRuleContext, n:int):
    """This node represents a virtual ParseTree node that does not come from
    the parser but rather is constructed via a rewriting rule. For example,
    the expression "a-b" is parsed as "Plus[a, Times[-1, b]]" in exitPlusOp(),
    so a node for "-1" needs to be created even though "-1" does not appear as
    a token in the token stream.

    Note that makeNumber sets the NumberContext's parent but does not add
    anything to parent's children."""

    # The hierarchy is:
    #   CommonToken->TerminalNodeImpl->NumberLiteralContext
    #       ->NumberBaseTenContext->NumberContext->parent

    digits_token = CommonToken(type=InputFormParser.DIGITS)
    digits_token._text = str(n)

    number = InputFormParser.NumberContext(None, InputFormParser.ExprContext(None, parent=parent))

    number_literal = InputFormParser.NumberBaseTenContext(None, InputFormParser.NumberLiteralContext(None, parent=number))

    number_literal.addTokenNode(digits_token)
    addChild(number, number_literal)

    return number


def flatten(ctx:ParserRuleContext):
    """Takes a ParserRuleContext of a binary operator and "flattens"
    the operator if one of its operands is the same binary operator context.

    This function only flattens if the operator is the same and also
    keeps the operators intact."""

    # If the child isn't the same construct, nothing to do.
    if ctx.getChild(0).__class__ != ctx.__class__:
        return

    lhs = ctx.getChild(0)
    op = ctx.getChild(1)
    rhs = ctx.getChild(2)
    lhsop = lhs.getChild(1)

    # If the operator of the nested Context isn't the same, nothing to do.
    # The operator is always in position 1 for infix operators. We do this
    # check because some Contexts that use the same context for multiple
    # operators.
    if op.getSymbol().type != lhsop.getSymbol().type:
        return

    # Clear all children.
    ctx.children.clear()
    # Add all children of lhs. (Also adds the operator of the lhs.)
    ctx.children.extend(lhs.children)
    # Finally, add the rhs back in.
    ctx.children.append(rhs)


def rewriteSpan(ctx:ParserRuleContext):
    opIndex = [i for (i, c) in enumerate(ctx.children) if c.getText() == ';;']
    spanExpressions = []

    opIndexSize = len(opIndex)
    childrenSize = len(ctx.children)
    # We emulate a for-loop with a while-loop. We can't use python's for
    # because we need to change the index variables in the loop body.
    i = 0
    nextOp = 0
    while True:
        if nextOp + 1 < opIndexSize \
            and opIndex[nextOp + 1] + 1 < childrenSize \
                and isinstance(ctx.children[opIndex[nextOp + 1] + 1], InputFormParser.ExprContext):
            # There is a next ";;"
            # and there is a node after the next ";;"
            # and there is a second ";;" followed by an expr.
            i = opIndex[nextOp + 1] + 1
            spanExpressions.append(i)
            # We want nextOp to end at the last ";;" of the current expression.
            nextOp += 1
        else:
            # There is no second ";;" belonging to this expression.
            if opIndex[nextOp] + 1 < childrenSize \
                    and isinstance(ctx.children[opIndex[nextOp + 1] + 1], InputFormParser.ExprContext):
                # There is a node after ";;"
                # and this span expression ends in an expr.
                i = opIndex[nextOp] + 1
                spanExpressions.append(i)
            else:
                # This span expression ends in the current ";;".
                i = opIndex[nextOp]
                spanExpressions.append(i)
        # Check for end of for-loop
        if i >= childrenSize or nextOp >= opIndexSize:
            break
        # Set up the next loop.
        i += 1
        nextOp +=1

    # At this point spanExpressions holds the index of the last child of each
    # span expression. It might be that after all of this there is nothing to do.
    if len(spanExpressions) == 1:
        return
    # Otherwise there is more than one span expression, and we need to rewrite
    # the tree replacing the Span?Context this method was invoked on with a
    # TimesContext.
    timesctx = InputFormParser.TimesContext(None, ctx)
    timesctx.children = [] # Clear children

    # Add each span expression as a child to timesctx.
    j = 0
    for i, spanExp in enumerate(spanExpressions):
        # i is the index of the current span expression in spanExpressions,
        # and j is the index to the beginning of the new span expression's
        # children in ctx.children. We make new SpanAContext objects for each
        # span expression.
        span = InputFormParser.SpanAContext(None, ctx)
        adopt(timesctx, span)
        span.children = ctx.children[j:spanExpressions[i]]
        # update j to be the beginning of the next expression.
        j = spanExpressions[i] + 1

    # Finally, detach the span this method was invoked on from its parent and
    # replace with the TimesContext.
    parentsChildren = ctx.parentCtx.children
    ctxIndex = parentsChildren.index(ctx)
    parentsChildren[ctxIndex] = timesctx
    ctx.parentCtx = timesctx



class PostParser(InputFormListener):

    def exitComparison(self, ctx:InputFormParser.ComparisonContext):
        """Inequality[]

        This function flattens and keeps the operators intact. It differs from
        flatten() in that we flatten if the class is the same but don't check
        if the operator is the same."""

        # If the child isn't the same construct, nothing to do.
        if ctx.getChild(0).__class__ != ctx.__class__:
            return

        lhs = ctx.getChild(0)
        op = ctx.getChild(1)
        rhs = ctx.getChild(2)
        lhsop = lhs.getChild(1)

        # This is where we differ from flatten(). We don't do the following
        # check.
        # if op.getSymbol().getType() != lhsop.getSymbol().getType():
        #     return

        # Clear all children.
        ctx.children.clear()
        # Add all children of lhs. (Also adds the operator of the lhs.)
        ctx.children.extend(lhs.children)
        # Keep the operator.
        ctx.children.append(op)
        # Finally, add the rhs back in.
        ctx.children.append(rhs)

    def exitCompoundExpression(self, ctx:InputFormParser.CompoundExpressionContext):
        """Composition[expr1,expr2]	e@*e@*e.

        ANTLR4 parses this rule as right associative for some reason, so
        we cannot use flatten(). The code is actually much simpler than
        flatten because of the right associativity."""

        childCount = ctx.getChildCount()

        # If there is no RHS, nothing to do.
        if childCount < 3:
            return

        rhs = ctx.getChild(childCount-1)

        # If the RHS child isn't the same construct, nothing to do.
        if rhs.__class__ != ctx.__class__:
            return

        # Remove RHS child.
        ctx.removeLastChild()
        # Add all children of rhs. (Also adds the operator of the rhs.)
        ctx.children.extend(rhs.children)

    def exitComposition(self, ctx:InputFormParser.CompositionContext):
        flatten(ctx)

    def exitRightComposition(self, ctx:InputFormParser.RightCompositionContext):
        flatten(ctx)

    def exitStringJoin(self, ctx:InputFormParser.StringJoinContext):
        flatten(ctx)

    def exitSmallCircle(self, ctx:InputFormParser.SmallCircleContext):
        flatten(ctx)

    def exitSpan(self, ctx:InputFormParser.ExprContext):
        """Span[expr1,expr2,expr3] e;;e;;e
        Parsing Span nodes is a complete mess."""

        # The only difference between the code for exitSpanA and exitSpanB is
        # the index of the rhs.
        rhsIndex = 1
        if isinstance(ctx, InputFormParser.SpanBContext):
            rhsIndex = 0


        # If there is no RHS expr, nothing to do.
        if len(ctx.expr()) == rhsIndex:
            return

        # Get the RHS expr.
        rhs = ctx.expr(rhsIndex)

        # If the RHS child isn't a SpanA, nothing to do.
        if not isinstance(rhs, InputFormParser.SpanAContext):
            return

        # Remove the last expr
        ctx.removeLastChild()
        # Replace it with its children
        ctx.children.extend(rhs.children)

        # If this is the topmost Span context, rewrite the tree.
        if not (isinstance(ctx.parentCtx, InputFormParser.SpanAContext)
                or isinstance(ctx.parentCtx, InputFormParser.SpanBContext)):
            rewriteSpan(ctx)

    def exitSpanA(self, ctx:InputFormParser.SpanAContext):
        """Span[expr1,expr2,expr3] e;;e;;e
        Parsing Span nodes is a complete mess."""
        self.exitSpan(ctx)

    def exitSpanB(self, ctx:InputFormParser.SpanBContext):
        """Span[expr1,expr2,expr3] e;;e;;e
        Parsing Span nodes is a complete mess."""
        self.exitSpan(ctx)

    def exitCircleDot(self, ctx:InputFormParser.CircleDotContext):
        flatten(ctx)

    def exitNonCommutativeMultiply(self, ctx:InputFormParser.NonCommutativeMultiplyContext):
        flatten(ctx)

    def exitCross(self, ctx:InputFormParser.CrossContext):
        flatten(ctx)

    def exitDot(self, ctx:InputFormParser.DotContext):
        flatten(ctx)

    def exitDiamond(self, ctx:InputFormParser.DiamondContext):
        flatten(ctx)

    def exitWedge(self, ctx:InputFormParser.WedgeContext):
        flatten(ctx)

    def exitVee(self, ctx:InputFormParser.VeeContext):
        flatten(ctx)

    def exitCircleTimes(self, ctx:InputFormParser.CircleTimesContext):
        flatten(ctx)

    def exitCenterDot(self, ctx:InputFormParser.CenterDotContext):
        flatten(ctx)

    def exitTimes(self, ctx:InputFormParser.TimesContext):
        """Times[expr1,expr2]

        We need to flatten over both Times and implicit Times. So
        flatten() isn't going to cut it because sometimes there is
        no operator."""

        # If the child isn't the same construct, nothing to do.
        if not isinstance(ctx.getChild(0), InputFormParser.TimesContext):
            return

        lhs = ctx.expr(0)
        rhs = ctx.expr(1)
        # Replace children with all children of lhs. (Also adds the operator of the lhs.)
        ctx.children = lhs.children
        # Add the rhs back in.
        ctx.children.append(rhs)

    def exitStar(self, ctx:InputFormParser.StarContext):
            flatten(ctx)

    def exitVerticalTilde(self, ctx:InputFormParser.VerticalTildeContext):
        flatten(ctx)

    def exitCoproduct(self, ctx:InputFormParser.CoproductContext):
        flatten(ctx)

    def exitCap(self, ctx:InputFormParser.CapContext):
        flatten(ctx)

    def exitCup(self, ctx:InputFormParser.CupContext):
        flatten(ctx)

    def exitCirclePlus(self, ctx:InputFormParser.CirclePlusContext):
        flatten(ctx)

    def exitPlusOp(self, ctx:InputFormParser.PlusOpContext):
        """PlusOp[expr1,expr2]

        We have to treat PlusOp special, because we have to keep the
        operators intact, and only  plus and minus (not PlusMinus or
        MinusPlus) are flat. The situation is complicated by the fact
        that Mathematica parses "a-b" as "Plus[a, Times[-1, b]]". We
        Rewrite the parse tree, inserting the Times context and
        changing BINARYMINUS to BINARYPLUS."""

        # If the op isn't Plus or Minus, nothing to do.
        if ctx.BINARYMINUS() is None and ctx.BINARYPLUS() is None:
            return

        # Since ANTLR4 parses this operator as left associative, we only
        # need to check the left hand side expr.
        rhs = ctx.getChild(2)

        # If the operator of the PlusOp is BINARYMINUS, we rewrite the tree as
        # "Plus[lhs, Times[-1, rhs]]". Note that if rhs is TIMES, we have to
        # keep that TIMES flat.
        if ctx.BINARYMINUS() is not None:
            # Construct Times, or obtain it from the rhs.
            times = None
            if isinstance(rhs, InputFormParser.TimesContext):
                times = rhs
            else:
                # If rhs is already a times, keep it flat.
                times = InputFormParser.TimesContext(None, InputFormParser.ExprContext(None))
                ctx.children.remove(rhs)
                adopt(ctx, times)
                adopt(times, rhs)
            # Add "-1" as the first child of Times.
            addChild(times, makeNumber(times, -1), 0)

            # Finally, we have to change operator to BINARYPLUS.
            plustoken = CommonToken(type=InputFormParser.BINARYPLUS)
            plustoken.text = '+'
            plus = TerminalNodeImpl(plustoken)
            # Replace minus token with plus.
            ctx.children[1] = plus
            plus.parentCtx = ctx

        # Flatten
        flatten(ctx)

    def exitIntersection(self, ctx:InputFormParser.IntersectionContext):
        flatten(ctx)

    def exitUnion(self, ctx:InputFormParser.UnionContext):
        flatten(ctx)

    def exitVerticalBar(self, ctx:InputFormParser.VerticalBarContext):
        flatten(ctx)

    def exitSame(self, ctx:InputFormParser.SameContext):
        flatten(ctx)

    def exitSetContainment(self, ctx:InputFormParser.SetContainmentContext):
        flatten(ctx)

    def exitAndOr(self, ctx:InputFormParser.ExprContext, opText:str):
        """Unifies the code for flattening And/Nand and Or/Nor."""

        # If the child isn't the same construct, nothing to do.
        if ctx.getChild(0).__class__ != ctx.__class__:
            return

        lhs = ctx.getChild(0)
        op = ctx.getChild(1)
        rhs = ctx.getChild(2)
        lhsop = lhs.getChild(1).getText()
        # Here's the part that's different from flatten().
        # If childOp is an Nand or parentOp is a Nand, then we need child==parent.
        if (lhsop == opText or op.getText() == opText) and op.getText() != lhsop:
            return

        # Clear all children.
        ctx.children.clear()
        # Add all children of lhs. (Also adds the operator of the lhs.)
        ctx.children.extend(lhs.children)
        # Finally, add the rhs back in.
        ctx.children.append(rhs)

    def exitAnd(self, ctx:InputFormParser.AndContext):
        self.exitAndOr(ctx, '\u22bc')

    def exitOr(self, ctx:InputFormParser.OrContext):
        self.exitAndOr(ctx, '\u22bd')

    def exitXor(self, ctx:InputFormParser.XorContext):
        flatten(ctx)

    def exitEquivalent(self, ctx:InputFormParser.EquivalentContext):
        flatten(ctx)

    def exitAlternatives(self, ctx:InputFormParser.AlternativesContext):
        flatten(ctx)

    def exitStringExpression(self, ctx:InputFormParser.StringExpressionContext):
        flatten(ctx)

    def exitColon(self, ctx:InputFormParser.ColonContext):
        flatten(ctx)

    def exitVerticalSeparator(self, ctx:InputFormParser.VerticalSeparatorContext):
        flatten(ctx)



from FoxySheep.tree.functions import (
    FunctionDeclarationNode, FunctionCallNode
    )

from FoxySheep.tree.nodes import treeNode, Symbol
from FoxySheep.tree.symbol_node import SymbolNode

from FoxySheep.tree.pretty_printer import PrettyPrinterCompact, PrettyPrinter, pretty_print, pretty_print_string

# Should we import all node classes, or should we only import the modules?



# Generated from FullForm.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .FullFormParser import FullFormParser
else:
    from FullFormParser import FullFormParser

# This class defines a complete generic visitor for a parse tree produced by FullFormParser.

class FullFormVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by FullFormParser#prog.
    def visitProg(self, ctx:FullFormParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#Number.
    def visitNumber(self, ctx:FullFormParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#StringLiteral.
    def visitStringLiteral(self, ctx:FullFormParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#SymbolLiteral.
    def visitSymbolLiteral(self, ctx:FullFormParser.SymbolLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#HeadExpression.
    def visitHeadExpression(self, ctx:FullFormParser.HeadExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#ExpressionListed.
    def visitExpressionListed(self, ctx:FullFormParser.ExpressionListedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#ContextName.
    def visitContextName(self, ctx:FullFormParser.ContextNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#SimpleContext.
    def visitSimpleContext(self, ctx:FullFormParser.SimpleContextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#CompoundContext.
    def visitCompoundContext(self, ctx:FullFormParser.CompoundContextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#NumberBaseN.
    def visitNumberBaseN(self, ctx:FullFormParser.NumberBaseNContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#NumberBaseTen.
    def visitNumberBaseTen(self, ctx:FullFormParser.NumberBaseTenContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#numberLiteralPrecision.
    def visitNumberLiteralPrecision(self, ctx:FullFormParser.NumberLiteralPrecisionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by FullFormParser#numberLiteralExponent.
    def visitNumberLiteralExponent(self, ctx:FullFormParser.NumberLiteralExponentContext):
        return self.visitChildren(ctx)



del FullFormParser


import astor
from antlr4 import TerminalNode
from antlr4.ParserRuleContext import ParserRuleContext
from FoxySheep.generated.InputFormVisitor import InputFormVisitor
from FoxySheep.transform.if_transform import input_form_post
import ast

IF_name_to_pyop = {
    "DivideContext": ast.Div,
    "MinusOpContext": ast.Sub,
    "TimesContext": ast.Mult,
    "PowerContext": ast.Pow,
    "NonCommutativeMultiplyContext": ast.Mult,  # Not quite right: doesn't capture non-commutativenewss
    "PlusOpContext": ast.Add,
}

fn_translate = {
    "GCD": "math.gcd",
    "Sin": "math.sin",
    "Plot": "matplotlib.pyplot.plot",
}


class InputForm2PyAst(InputFormVisitor):
    def get_full_form(self, e) -> ast.AST:
        """
        Returns a Python AST for the ParseTree `e`
        """
        if isinstance(e, TerminalNode):
            return e.getText()

        return self.visit(e)

    def get_child_list(self, efunction) -> ast.AST:
        """
        Convert antlr4's awkward function-call API to a list.

        We assume efunction is a function of an integer i. This awkward
        interface is from antlr4's Python API.
        """

        elist = []
        i = 0
        e = efunction(i)
        while e is not None:
            elist.append(e)
            i += 1
            e = efunction(i)
        return elist

    def visitProg(self, ctx: ParserRuleContext) -> ast.AST:
        exprs = ctx.expr()
        n = len(exprs)
        if n > 0:
            if n == 1:
                return self.visit(exprs[0])
            else:
                expr_list = map(self.visit, exprs)
                pass
            pass
        elif n == 0:
            expr_list = []
            for expr in ctx.expressionList().getChildren():
                if expr.getText() == ",":
                    continue
                expr_list.append(self.visit(expr))
        return ast.Tuple(expr_list, 5)

    def visitList(self, ctx:ParserRuleContext) -> ast.AST:
        node = ast.List(expr_ctx="Load()")
        expr_list = []
        for expr in ctx.expressionList().getChildren():
            if expr.getText() == ",":
                continue
            expr_list.append(self.visit(expr))
        node.elts = expr_list
        return node

    def visitParentheses(self, ctx:ParserRuleContext) -> ast.AST:
        return self.visit(ctx.getChild(1))

    def visitNumberBaseTen(self, ctx: ParserRuleContext) -> ast.AST:
        digits = ctx.getText()
        if digits.find(".") >= 0:
            ast_top = ast.parse(f"decimal.Decimal({digits})")
            node = ast_top.body[0]
        else:
            if digits.endswith("`"):
                # Python doesn't have machine-specific representation. So drop
                # off the indicator.
                digits = digits[:-1]
            node = ast.Constant()
            node.lineno = 0
            node.col_offset = 0
            node.value = int(digits, 10)
        return node

    def visitNumberBaseN(self, ctx: ParserRuleContext) -> ast.AST:
        digits = ctx.getChild(0).getText()
        base = ctx.getChild(1).getText()
        if base.startswith("^^"):
            ast_top = ast.parse(f"int({digits}, {base[2:]})")
            node = ast_top.body[0]
        else:
            node = ast.Constant()
            node.lineno = 0
            node.col_offset = 0
            node.value = int(ctx.getText(), 10)
        return node

    def visitOutNumbered(self, ctx: ParserRuleContext) -> ast.AST:
        fn_name_node = ast.Name(id="Out", ctx="Load()")
        args = [ast.Constant(number)]
        return ast.Call(func=fn_name_node, args=args, keywords=[])

    def visitOutUnnumbered(self, ctx: ParserRuleContext) -> ast.AST:
        fn_name_node = ast.Name(id="Out", ctx="Load()")
        return ast.Call(func=fn_name_node, args=[], keywords=[])

    def visitHeadExpression(self, ctx: ParserRuleContext) -> ast.AST:
        "Translates function calls"
        fn_name = ctx.expr().getText()
        fn_name = fn_translate.get(fn_name, fn_name)
        fn_name_node = ast.Name(id=fn_name, ctx="Load()")
        args = []
        for arg in ctx.expressionList().getChildren():
            if arg.getText() == ",":
                continue
            args.append(self.visit(arg))
        return ast.Call(func=fn_name_node, args=args, keywords=[])

        # return self.make_head(self.get_full_form(ctx.expr()), ctx.expressionList())
        return None

    def visitTimes(self, ctx: ParserRuleContext) -> ast.AST:
        """
        Translates infix multiplcation.
        """
        node = ast.BinOp()
        ctx_name = type(ctx).__name__
        ast_op_fn = IF_name_to_pyop.get(ctx_name, None)
        if ast_op_fn:
            node.op = ast_op_fn()
        else:
            raise RuntimeError(f"Unknown op context {type(ctx_name)}")

        node.left = self.visit(ctx.expr(0))
        node.right = self.visit(ctx.expr(1))
        return node

    def visitPlusOp(self, ctx: ParserRuleContext) -> ast.AST:
        """
        Translates infix binary operators. Function binary operators are different.
        """
        node = ast.BinOp()
        if ctx.BINARYMINUS():
            node.op = ast.Sub()
        elif ctx.BINARYPLUS():
            node.op = ast.Add()
        elif ctx.BINARYMINUSPLUS() or ctx.BINARYPLUSMINUS():
            raise RuntimeError(f"Can't handle +- or -+")
        else:
            raise RuntimeError(f"Unknown op context {ctx}")

        node.left = self.visit(ctx.expr(0))
        node.right = self.visit(ctx.expr(1))
        return node

    def visitPower(self, ctx: ParserRuleContext) -> ast.AST:
        """
        Translates infix binary operators. Function binary operators are different.
        """
        node = ast.BinOp()
        ctx_name = type(ctx).__name__
        ast_op_fn = IF_name_to_pyop.get(ctx_name, None)
        if ast_op_fn:
            node.op = ast_op_fn()
        else:
            raise RuntimeError(f"Unknown op context {type(ctx_name)}")

        node.left = self.visit(ctx.expr(0))
        node.right = self.visit(ctx.expr(1))
        return node

    visitMinusOp = (
        visitDivide
    ) = visitNonCommutativeMultiply = visitPower


    def visitUnaryPlusMinus(self, ctx: ParserRuleContext) -> ast.AST:
        """Translates prefix + and - operators. Note +- (plus or minus) and -+
        are different.
        """
        node = ast.UnaryOp()
        node.lineno = 0
        node.col_offset = 0
        ctx_name = type(ctx).__name__

        if ctx.MINUS() is not None:
            node.op = ast.USub()
        elif ctx.PLUS() is not None:
            node.op = ast.UAdd()
        elif ctx.PLUSMINUS() or ctx.MINUSPLUS():
            # We don't handle these
            return self.visitChildren(ctx)
        node.operand = self.visit(ctx.expr())
        return node

    def visitSymbolLiteral(self, ctx:ParserRuleContext) -> ast.AST:
        return ast.Name(ctx.getText())


def input_form_to_python_ast(tree) -> ast.AST:
    transform = InputForm2PyAst()
    return transform.visit(tree)


def input_form_to_python(input_form_str: str, parse_tree_fn, show_tree_fn=None) -> str:

    tree = parse_tree_fn(input_form_str, show_tree_fn=show_tree_fn)
    pyast = input_form_to_python_ast(tree)
    return astor.to_source(pyast)


if __name__ == "__main__":

    def parse_tree_fn(expr: str, show_tree_fn):
        from FoxySheep.generated.InputFormLexer import InputFormLexer
        from FoxySheep.generated.InputFormParser import InputFormParser
        from antlr4 import InputStream, CommonTokenStream

        lexer = InputFormLexer(InputStream(expr))
        parser = InputFormParser(CommonTokenStream(lexer))
        tree = parser.prog()
        show_tree_fn(tree, parser.ruleNames)
        return tree

    from FoxySheep.tree.pretty_printer import pretty_print_compact

    print(input_form_to_python("1 + 2 - 3 + 4", parse_tree_fn, pretty_print_compact))



# Generated from InputForm.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .InputFormParser import InputFormParser
else:
    from InputFormParser import InputFormParser

# This class defines a complete listener for a parse tree produced by InputFormParser.
class InputFormListener(ParseTreeListener):

    # Enter a parse tree produced by InputFormParser#prog.
    def enterProg(self, ctx:InputFormParser.ProgContext):
        pass

    # Exit a parse tree produced by InputFormParser#prog.
    def exitProg(self, ctx:InputFormParser.ProgContext):
        pass


    # Enter a parse tree produced by InputFormParser#PatternExp.
    def enterPatternExp(self, ctx:InputFormParser.PatternExpContext):
        pass

    # Exit a parse tree produced by InputFormParser#PatternExp.
    def exitPatternExp(self, ctx:InputFormParser.PatternExpContext):
        pass


    # Enter a parse tree produced by InputFormParser#Or.
    def enterOr(self, ctx:InputFormParser.OrContext):
        pass

    # Exit a parse tree produced by InputFormParser#Or.
    def exitOr(self, ctx:InputFormParser.OrContext):
        pass


    # Enter a parse tree produced by InputFormParser#Conjugate.
    def enterConjugate(self, ctx:InputFormParser.ConjugateContext):
        pass

    # Exit a parse tree produced by InputFormParser#Conjugate.
    def exitConjugate(self, ctx:InputFormParser.ConjugateContext):
        pass


    # Enter a parse tree produced by InputFormParser#Ceiling.
    def enterCeiling(self, ctx:InputFormParser.CeilingContext):
        pass

    # Exit a parse tree produced by InputFormParser#Ceiling.
    def exitCeiling(self, ctx:InputFormParser.CeilingContext):
        pass


    # Enter a parse tree produced by InputFormParser#Infix.
    def enterInfix(self, ctx:InputFormParser.InfixContext):
        pass

    # Exit a parse tree produced by InputFormParser#Infix.
    def exitInfix(self, ctx:InputFormParser.InfixContext):
        pass


    # Enter a parse tree produced by InputFormParser#Therefore.
    def enterTherefore(self, ctx:InputFormParser.ThereforeContext):
        pass

    # Exit a parse tree produced by InputFormParser#Therefore.
    def exitTherefore(self, ctx:InputFormParser.ThereforeContext):
        pass


    # Enter a parse tree produced by InputFormParser#TagUnset.
    def enterTagUnset(self, ctx:InputFormParser.TagUnsetContext):
        pass

    # Exit a parse tree produced by InputFormParser#TagUnset.
    def exitTagUnset(self, ctx:InputFormParser.TagUnsetContext):
        pass


    # Enter a parse tree produced by InputFormParser#Accessor.
    def enterAccessor(self, ctx:InputFormParser.AccessorContext):
        pass

    # Exit a parse tree produced by InputFormParser#Accessor.
    def exitAccessor(self, ctx:InputFormParser.AccessorContext):
        pass


    # Enter a parse tree produced by InputFormParser#CircleMinus.
    def enterCircleMinus(self, ctx:InputFormParser.CircleMinusContext):
        pass

    # Exit a parse tree produced by InputFormParser#CircleMinus.
    def exitCircleMinus(self, ctx:InputFormParser.CircleMinusContext):
        pass


    # Enter a parse tree produced by InputFormParser#Divide.
    def enterDivide(self, ctx:InputFormParser.DivideContext):
        pass

    # Exit a parse tree produced by InputFormParser#Divide.
    def exitDivide(self, ctx:InputFormParser.DivideContext):
        pass


    # Enter a parse tree produced by InputFormParser#Implies.
    def enterImplies(self, ctx:InputFormParser.ImpliesContext):
        pass

    # Exit a parse tree produced by InputFormParser#Implies.
    def exitImplies(self, ctx:InputFormParser.ImpliesContext):
        pass


    # Enter a parse tree produced by InputFormParser#PlusOp.
    def enterPlusOp(self, ctx:InputFormParser.PlusOpContext):
        pass

    # Exit a parse tree produced by InputFormParser#PlusOp.
    def exitPlusOp(self, ctx:InputFormParser.PlusOpContext):
        pass


    # Enter a parse tree produced by InputFormParser#RightComposition.
    def enterRightComposition(self, ctx:InputFormParser.RightCompositionContext):
        pass

    # Exit a parse tree produced by InputFormParser#RightComposition.
    def exitRightComposition(self, ctx:InputFormParser.RightCompositionContext):
        pass


    # Enter a parse tree produced by InputFormParser#NonCommutativeMultiply.
    def enterNonCommutativeMultiply(self, ctx:InputFormParser.NonCommutativeMultiplyContext):
        pass

    # Exit a parse tree produced by InputFormParser#NonCommutativeMultiply.
    def exitNonCommutativeMultiply(self, ctx:InputFormParser.NonCommutativeMultiplyContext):
        pass


    # Enter a parse tree produced by InputFormParser#List.
    def enterList(self, ctx:InputFormParser.ListContext):
        pass

    # Exit a parse tree produced by InputFormParser#List.
    def exitList(self, ctx:InputFormParser.ListContext):
        pass


    # Enter a parse tree produced by InputFormParser#Cup.
    def enterCup(self, ctx:InputFormParser.CupContext):
        pass

    # Exit a parse tree produced by InputFormParser#Cup.
    def exitCup(self, ctx:InputFormParser.CupContext):
        pass


    # Enter a parse tree produced by InputFormParser#Same.
    def enterSame(self, ctx:InputFormParser.SameContext):
        pass

    # Exit a parse tree produced by InputFormParser#Same.
    def exitSame(self, ctx:InputFormParser.SameContext):
        pass


    # Enter a parse tree produced by InputFormParser#Optional.
    def enterOptional(self, ctx:InputFormParser.OptionalContext):
        pass

    # Exit a parse tree produced by InputFormParser#Optional.
    def exitOptional(self, ctx:InputFormParser.OptionalContext):
        pass


    # Enter a parse tree produced by InputFormParser#SuchThat.
    def enterSuchThat(self, ctx:InputFormParser.SuchThatContext):
        pass

    # Exit a parse tree produced by InputFormParser#SuchThat.
    def exitSuchThat(self, ctx:InputFormParser.SuchThatContext):
        pass


    # Enter a parse tree produced by InputFormParser#DoubleBracketingBar.
    def enterDoubleBracketingBar(self, ctx:InputFormParser.DoubleBracketingBarContext):
        pass

    # Exit a parse tree produced by InputFormParser#DoubleBracketingBar.
    def exitDoubleBracketingBar(self, ctx:InputFormParser.DoubleBracketingBarContext):
        pass


    # Enter a parse tree produced by InputFormParser#PatternBlankDot.
    def enterPatternBlankDot(self, ctx:InputFormParser.PatternBlankDotContext):
        pass

    # Exit a parse tree produced by InputFormParser#PatternBlankDot.
    def exitPatternBlankDot(self, ctx:InputFormParser.PatternBlankDotContext):
        pass


    # Enter a parse tree produced by InputFormParser#Dot.
    def enterDot(self, ctx:InputFormParser.DotContext):
        pass

    # Exit a parse tree produced by InputFormParser#Dot.
    def exitDot(self, ctx:InputFormParser.DotContext):
        pass


    # Enter a parse tree produced by InputFormParser#VerticalBar.
    def enterVerticalBar(self, ctx:InputFormParser.VerticalBarContext):
        pass

    # Exit a parse tree produced by InputFormParser#VerticalBar.
    def exitVerticalBar(self, ctx:InputFormParser.VerticalBarContext):
        pass


    # Enter a parse tree produced by InputFormParser#Square.
    def enterSquare(self, ctx:InputFormParser.SquareContext):
        pass

    # Exit a parse tree produced by InputFormParser#Square.
    def exitSquare(self, ctx:InputFormParser.SquareContext):
        pass


    # Enter a parse tree produced by InputFormParser#Alternatives.
    def enterAlternatives(self, ctx:InputFormParser.AlternativesContext):
        pass

    # Exit a parse tree produced by InputFormParser#Alternatives.
    def exitAlternatives(self, ctx:InputFormParser.AlternativesContext):
        pass


    # Enter a parse tree produced by InputFormParser#Out.
    def enterOut(self, ctx:InputFormParser.OutContext):
        pass

    # Exit a parse tree produced by InputFormParser#Out.
    def exitOut(self, ctx:InputFormParser.OutContext):
        pass


    # Enter a parse tree produced by InputFormParser#BoxConstructor.
    def enterBoxConstructor(self, ctx:InputFormParser.BoxConstructorContext):
        pass

    # Exit a parse tree produced by InputFormParser#BoxConstructor.
    def exitBoxConstructor(self, ctx:InputFormParser.BoxConstructorContext):
        pass


    # Enter a parse tree produced by InputFormParser#Not.
    def enterNot(self, ctx:InputFormParser.NotContext):
        pass

    # Exit a parse tree produced by InputFormParser#Not.
    def exitNot(self, ctx:InputFormParser.NotContext):
        pass


    # Enter a parse tree produced by InputFormParser#Postfix.
    def enterPostfix(self, ctx:InputFormParser.PostfixContext):
        pass

    # Exit a parse tree produced by InputFormParser#Postfix.
    def exitPostfix(self, ctx:InputFormParser.PostfixContext):
        pass


    # Enter a parse tree produced by InputFormParser#PatternBlanks.
    def enterPatternBlanks(self, ctx:InputFormParser.PatternBlanksContext):
        pass

    # Exit a parse tree produced by InputFormParser#PatternBlanks.
    def exitPatternBlanks(self, ctx:InputFormParser.PatternBlanksContext):
        pass


    # Enter a parse tree produced by InputFormParser#UnaryPlusMinus.
    def enterUnaryPlusMinus(self, ctx:InputFormParser.UnaryPlusMinusContext):
        pass

    # Exit a parse tree produced by InputFormParser#UnaryPlusMinus.
    def exitUnaryPlusMinus(self, ctx:InputFormParser.UnaryPlusMinusContext):
        pass


    # Enter a parse tree produced by InputFormParser#Cap.
    def enterCap(self, ctx:InputFormParser.CapContext):
        pass

    # Exit a parse tree produced by InputFormParser#Cap.
    def exitCap(self, ctx:InputFormParser.CapContext):
        pass


    # Enter a parse tree produced by InputFormParser#CirclePlus.
    def enterCirclePlus(self, ctx:InputFormParser.CirclePlusContext):
        pass

    # Exit a parse tree produced by InputFormParser#CirclePlus.
    def exitCirclePlus(self, ctx:InputFormParser.CirclePlusContext):
        pass


    # Enter a parse tree produced by InputFormParser#Because.
    def enterBecause(self, ctx:InputFormParser.BecauseContext):
        pass

    # Exit a parse tree produced by InputFormParser#Because.
    def exitBecause(self, ctx:InputFormParser.BecauseContext):
        pass


    # Enter a parse tree produced by InputFormParser#StringLiteral.
    def enterStringLiteral(self, ctx:InputFormParser.StringLiteralContext):
        pass

    # Exit a parse tree produced by InputFormParser#StringLiteral.
    def exitStringLiteral(self, ctx:InputFormParser.StringLiteralContext):
        pass


    # Enter a parse tree produced by InputFormParser#And.
    def enterAnd(self, ctx:InputFormParser.AndContext):
        pass

    # Exit a parse tree produced by InputFormParser#And.
    def exitAnd(self, ctx:InputFormParser.AndContext):
        pass


    # Enter a parse tree produced by InputFormParser#Get.
    def enterGet(self, ctx:InputFormParser.GetContext):
        pass

    # Exit a parse tree produced by InputFormParser#Get.
    def exitGet(self, ctx:InputFormParser.GetContext):
        pass


    # Enter a parse tree produced by InputFormParser#Equivalent.
    def enterEquivalent(self, ctx:InputFormParser.EquivalentContext):
        pass

    # Exit a parse tree produced by InputFormParser#Equivalent.
    def exitEquivalent(self, ctx:InputFormParser.EquivalentContext):
        pass


    # Enter a parse tree produced by InputFormParser#CompoundExpression.
    def enterCompoundExpression(self, ctx:InputFormParser.CompoundExpressionContext):
        pass

    # Exit a parse tree produced by InputFormParser#CompoundExpression.
    def exitCompoundExpression(self, ctx:InputFormParser.CompoundExpressionContext):
        pass


    # Enter a parse tree produced by InputFormParser#Derivative.
    def enterDerivative(self, ctx:InputFormParser.DerivativeContext):
        pass

    # Exit a parse tree produced by InputFormParser#Derivative.
    def exitDerivative(self, ctx:InputFormParser.DerivativeContext):
        pass


    # Enter a parse tree produced by InputFormParser#Slot.
    def enterSlot(self, ctx:InputFormParser.SlotContext):
        pass

    # Exit a parse tree produced by InputFormParser#Slot.
    def exitSlot(self, ctx:InputFormParser.SlotContext):
        pass


    # Enter a parse tree produced by InputFormParser#RightTee.
    def enterRightTee(self, ctx:InputFormParser.RightTeeContext):
        pass

    # Exit a parse tree produced by InputFormParser#RightTee.
    def exitRightTee(self, ctx:InputFormParser.RightTeeContext):
        pass


    # Enter a parse tree produced by InputFormParser#Xor.
    def enterXor(self, ctx:InputFormParser.XorContext):
        pass

    # Exit a parse tree produced by InputFormParser#Xor.
    def exitXor(self, ctx:InputFormParser.XorContext):
        pass


    # Enter a parse tree produced by InputFormParser#Rule.
    def enterRule(self, ctx:InputFormParser.RuleContext):
        pass

    # Exit a parse tree produced by InputFormParser#Rule.
    def exitRule(self, ctx:InputFormParser.RuleContext):
        pass


    # Enter a parse tree produced by InputFormParser#HeadExpression.
    def enterHeadExpression(self, ctx:InputFormParser.HeadExpressionContext):
        pass

    # Exit a parse tree produced by InputFormParser#HeadExpression.
    def exitHeadExpression(self, ctx:InputFormParser.HeadExpressionContext):
        pass


    # Enter a parse tree produced by InputFormParser#ReplaceAll.
    def enterReplaceAll(self, ctx:InputFormParser.ReplaceAllContext):
        pass

    # Exit a parse tree produced by InputFormParser#ReplaceAll.
    def exitReplaceAll(self, ctx:InputFormParser.ReplaceAllContext):
        pass


    # Enter a parse tree produced by InputFormParser#Intersection.
    def enterIntersection(self, ctx:InputFormParser.IntersectionContext):
        pass

    # Exit a parse tree produced by InputFormParser#Intersection.
    def exitIntersection(self, ctx:InputFormParser.IntersectionContext):
        pass


    # Enter a parse tree produced by InputFormParser#PreIncrement.
    def enterPreIncrement(self, ctx:InputFormParser.PreIncrementContext):
        pass

    # Exit a parse tree produced by InputFormParser#PreIncrement.
    def exitPreIncrement(self, ctx:InputFormParser.PreIncrementContext):
        pass


    # Enter a parse tree produced by InputFormParser#Integrate.
    def enterIntegrate(self, ctx:InputFormParser.IntegrateContext):
        pass

    # Exit a parse tree produced by InputFormParser#Integrate.
    def exitIntegrate(self, ctx:InputFormParser.IntegrateContext):
        pass


    # Enter a parse tree produced by InputFormParser#Set.
    def enterSet(self, ctx:InputFormParser.SetContext):
        pass

    # Exit a parse tree produced by InputFormParser#Set.
    def exitSet(self, ctx:InputFormParser.SetContext):
        pass


    # Enter a parse tree produced by InputFormParser#OpEquals.
    def enterOpEquals(self, ctx:InputFormParser.OpEqualsContext):
        pass

    # Exit a parse tree produced by InputFormParser#OpEquals.
    def exitOpEquals(self, ctx:InputFormParser.OpEqualsContext):
        pass


    # Enter a parse tree produced by InputFormParser#Message.
    def enterMessage(self, ctx:InputFormParser.MessageContext):
        pass

    # Exit a parse tree produced by InputFormParser#Message.
    def exitMessage(self, ctx:InputFormParser.MessageContext):
        pass


    # Enter a parse tree produced by InputFormParser#Cross.
    def enterCross(self, ctx:InputFormParser.CrossContext):
        pass

    # Exit a parse tree produced by InputFormParser#Cross.
    def exitCross(self, ctx:InputFormParser.CrossContext):
        pass


    # Enter a parse tree produced by InputFormParser#PatternTest.
    def enterPatternTest(self, ctx:InputFormParser.PatternTestContext):
        pass

    # Exit a parse tree produced by InputFormParser#PatternTest.
    def exitPatternTest(self, ctx:InputFormParser.PatternTestContext):
        pass


    # Enter a parse tree produced by InputFormParser#Prefix.
    def enterPrefix(self, ctx:InputFormParser.PrefixContext):
        pass

    # Exit a parse tree produced by InputFormParser#Prefix.
    def exitPrefix(self, ctx:InputFormParser.PrefixContext):
        pass


    # Enter a parse tree produced by InputFormParser#Backslash.
    def enterBackslash(self, ctx:InputFormParser.BackslashContext):
        pass

    # Exit a parse tree produced by InputFormParser#Backslash.
    def exitBackslash(self, ctx:InputFormParser.BackslashContext):
        pass


    # Enter a parse tree produced by InputFormParser#Repeated.
    def enterRepeated(self, ctx:InputFormParser.RepeatedContext):
        pass

    # Exit a parse tree produced by InputFormParser#Repeated.
    def exitRepeated(self, ctx:InputFormParser.RepeatedContext):
        pass


    # Enter a parse tree produced by InputFormParser#MapApply.
    def enterMapApply(self, ctx:InputFormParser.MapApplyContext):
        pass

    # Exit a parse tree produced by InputFormParser#MapApply.
    def exitMapApply(self, ctx:InputFormParser.MapApplyContext):
        pass


    # Enter a parse tree produced by InputFormParser#Union.
    def enterUnion(self, ctx:InputFormParser.UnionContext):
        pass

    # Exit a parse tree produced by InputFormParser#Union.
    def exitUnion(self, ctx:InputFormParser.UnionContext):
        pass


    # Enter a parse tree produced by InputFormParser#VerticalSeparator.
    def enterVerticalSeparator(self, ctx:InputFormParser.VerticalSeparatorContext):
        pass

    # Exit a parse tree produced by InputFormParser#VerticalSeparator.
    def exitVerticalSeparator(self, ctx:InputFormParser.VerticalSeparatorContext):
        pass


    # Enter a parse tree produced by InputFormParser#Factorial.
    def enterFactorial(self, ctx:InputFormParser.FactorialContext):
        pass

    # Exit a parse tree produced by InputFormParser#Factorial.
    def exitFactorial(self, ctx:InputFormParser.FactorialContext):
        pass


    # Enter a parse tree produced by InputFormParser#SpanA.
    def enterSpanA(self, ctx:InputFormParser.SpanAContext):
        pass

    # Exit a parse tree produced by InputFormParser#SpanA.
    def exitSpanA(self, ctx:InputFormParser.SpanAContext):
        pass


    # Enter a parse tree produced by InputFormParser#Function.
    def enterFunction(self, ctx:InputFormParser.FunctionContext):
        pass

    # Exit a parse tree produced by InputFormParser#Function.
    def exitFunction(self, ctx:InputFormParser.FunctionContext):
        pass


    # Enter a parse tree produced by InputFormParser#Number.
    def enterNumber(self, ctx:InputFormParser.NumberContext):
        pass

    # Exit a parse tree produced by InputFormParser#Number.
    def exitNumber(self, ctx:InputFormParser.NumberContext):
        pass


    # Enter a parse tree produced by InputFormParser#Star.
    def enterStar(self, ctx:InputFormParser.StarContext):
        pass

    # Exit a parse tree produced by InputFormParser#Star.
    def exitStar(self, ctx:InputFormParser.StarContext):
        pass


    # Enter a parse tree produced by InputFormParser#Comparison.
    def enterComparison(self, ctx:InputFormParser.ComparisonContext):
        pass

    # Exit a parse tree produced by InputFormParser#Comparison.
    def exitComparison(self, ctx:InputFormParser.ComparisonContext):
        pass


    # Enter a parse tree produced by InputFormParser#TagSet.
    def enterTagSet(self, ctx:InputFormParser.TagSetContext):
        pass

    # Exit a parse tree produced by InputFormParser#TagSet.
    def exitTagSet(self, ctx:InputFormParser.TagSetContext):
        pass


    # Enter a parse tree produced by InputFormParser#Increment.
    def enterIncrement(self, ctx:InputFormParser.IncrementContext):
        pass

    # Exit a parse tree produced by InputFormParser#Increment.
    def exitIncrement(self, ctx:InputFormParser.IncrementContext):
        pass


    # Enter a parse tree produced by InputFormParser#VerticalTilde.
    def enterVerticalTilde(self, ctx:InputFormParser.VerticalTildeContext):
        pass

    # Exit a parse tree produced by InputFormParser#VerticalTilde.
    def exitVerticalTilde(self, ctx:InputFormParser.VerticalTildeContext):
        pass


    # Enter a parse tree produced by InputFormParser#Colon.
    def enterColon(self, ctx:InputFormParser.ColonContext):
        pass

    # Exit a parse tree produced by InputFormParser#Colon.
    def exitColon(self, ctx:InputFormParser.ColonContext):
        pass


    # Enter a parse tree produced by InputFormParser#SmallCircle.
    def enterSmallCircle(self, ctx:InputFormParser.SmallCircleContext):
        pass

    # Exit a parse tree produced by InputFormParser#SmallCircle.
    def exitSmallCircle(self, ctx:InputFormParser.SmallCircleContext):
        pass


    # Enter a parse tree produced by InputFormParser#Parentheses.
    def enterParentheses(self, ctx:InputFormParser.ParenthesesContext):
        pass

    # Exit a parse tree produced by InputFormParser#Parentheses.
    def exitParentheses(self, ctx:InputFormParser.ParenthesesContext):
        pass


    # Enter a parse tree produced by InputFormParser#SpanB.
    def enterSpanB(self, ctx:InputFormParser.SpanBContext):
        pass

    # Exit a parse tree produced by InputFormParser#SpanB.
    def exitSpanB(self, ctx:InputFormParser.SpanBContext):
        pass


    # Enter a parse tree produced by InputFormParser#Condition.
    def enterCondition(self, ctx:InputFormParser.ConditionContext):
        pass

    # Exit a parse tree produced by InputFormParser#Condition.
    def exitCondition(self, ctx:InputFormParser.ConditionContext):
        pass


    # Enter a parse tree produced by InputFormParser#Floor.
    def enterFloor(self, ctx:InputFormParser.FloorContext):
        pass

    # Exit a parse tree produced by InputFormParser#Floor.
    def exitFloor(self, ctx:InputFormParser.FloorContext):
        pass


    # Enter a parse tree produced by InputFormParser#Composition.
    def enterComposition(self, ctx:InputFormParser.CompositionContext):
        pass

    # Exit a parse tree produced by InputFormParser#Composition.
    def exitComposition(self, ctx:InputFormParser.CompositionContext):
        pass


    # Enter a parse tree produced by InputFormParser#CircleDot.
    def enterCircleDot(self, ctx:InputFormParser.CircleDotContext):
        pass

    # Exit a parse tree produced by InputFormParser#CircleDot.
    def exitCircleDot(self, ctx:InputFormParser.CircleDotContext):
        pass


    # Enter a parse tree produced by InputFormParser#SymbolLiteral.
    def enterSymbolLiteral(self, ctx:InputFormParser.SymbolLiteralContext):
        pass

    # Exit a parse tree produced by InputFormParser#SymbolLiteral.
    def exitSymbolLiteral(self, ctx:InputFormParser.SymbolLiteralContext):
        pass


    # Enter a parse tree produced by InputFormParser#CircleTimes.
    def enterCircleTimes(self, ctx:InputFormParser.CircleTimesContext):
        pass

    # Exit a parse tree produced by InputFormParser#CircleTimes.
    def exitCircleTimes(self, ctx:InputFormParser.CircleTimesContext):
        pass


    # Enter a parse tree produced by InputFormParser#Unset.
    def enterUnset(self, ctx:InputFormParser.UnsetContext):
        pass

    # Exit a parse tree produced by InputFormParser#Unset.
    def exitUnset(self, ctx:InputFormParser.UnsetContext):
        pass


    # Enter a parse tree produced by InputFormParser#Del.
    def enterDel(self, ctx:InputFormParser.DelContext):
        pass

    # Exit a parse tree produced by InputFormParser#Del.
    def exitDel(self, ctx:InputFormParser.DelContext):
        pass


    # Enter a parse tree produced by InputFormParser#Diamond.
    def enterDiamond(self, ctx:InputFormParser.DiamondContext):
        pass

    # Exit a parse tree produced by InputFormParser#Diamond.
    def exitDiamond(self, ctx:InputFormParser.DiamondContext):
        pass


    # Enter a parse tree produced by InputFormParser#Wedge.
    def enterWedge(self, ctx:InputFormParser.WedgeContext):
        pass

    # Exit a parse tree produced by InputFormParser#Wedge.
    def exitWedge(self, ctx:InputFormParser.WedgeContext):
        pass


    # Enter a parse tree produced by InputFormParser#Put.
    def enterPut(self, ctx:InputFormParser.PutContext):
        pass

    # Exit a parse tree produced by InputFormParser#Put.
    def exitPut(self, ctx:InputFormParser.PutContext):
        pass


    # Enter a parse tree produced by InputFormParser#StringJoin.
    def enterStringJoin(self, ctx:InputFormParser.StringJoinContext):
        pass

    # Exit a parse tree produced by InputFormParser#StringJoin.
    def exitStringJoin(self, ctx:InputFormParser.StringJoinContext):
        pass


    # Enter a parse tree produced by InputFormParser#Tee.
    def enterTee(self, ctx:InputFormParser.TeeContext):
        pass

    # Exit a parse tree produced by InputFormParser#Tee.
    def exitTee(self, ctx:InputFormParser.TeeContext):
        pass


    # Enter a parse tree produced by InputFormParser#SetContainment.
    def enterSetContainment(self, ctx:InputFormParser.SetContainmentContext):
        pass

    # Exit a parse tree produced by InputFormParser#SetContainment.
    def exitSetContainment(self, ctx:InputFormParser.SetContainmentContext):
        pass


    # Enter a parse tree produced by InputFormParser#Vee.
    def enterVee(self, ctx:InputFormParser.VeeContext):
        pass

    # Exit a parse tree produced by InputFormParser#Vee.
    def exitVee(self, ctx:InputFormParser.VeeContext):
        pass


    # Enter a parse tree produced by InputFormParser#CenterDot.
    def enterCenterDot(self, ctx:InputFormParser.CenterDotContext):
        pass

    # Exit a parse tree produced by InputFormParser#CenterDot.
    def exitCenterDot(self, ctx:InputFormParser.CenterDotContext):
        pass


    # Enter a parse tree produced by InputFormParser#Times.
    def enterTimes(self, ctx:InputFormParser.TimesContext):
        pass

    # Exit a parse tree produced by InputFormParser#Times.
    def exitTimes(self, ctx:InputFormParser.TimesContext):
        pass


    # Enter a parse tree produced by InputFormParser#StringExpression.
    def enterStringExpression(self, ctx:InputFormParser.StringExpressionContext):
        pass

    # Exit a parse tree produced by InputFormParser#StringExpression.
    def exitStringExpression(self, ctx:InputFormParser.StringExpressionContext):
        pass


    # Enter a parse tree produced by InputFormParser#BracketingBar.
    def enterBracketingBar(self, ctx:InputFormParser.BracketingBarContext):
        pass

    # Exit a parse tree produced by InputFormParser#BracketingBar.
    def exitBracketingBar(self, ctx:InputFormParser.BracketingBarContext):
        pass


    # Enter a parse tree produced by InputFormParser#Coproduct.
    def enterCoproduct(self, ctx:InputFormParser.CoproductContext):
        pass

    # Exit a parse tree produced by InputFormParser#Coproduct.
    def exitCoproduct(self, ctx:InputFormParser.CoproductContext):
        pass


    # Enter a parse tree produced by InputFormParser#AngleBracket.
    def enterAngleBracket(self, ctx:InputFormParser.AngleBracketContext):
        pass

    # Exit a parse tree produced by InputFormParser#AngleBracket.
    def exitAngleBracket(self, ctx:InputFormParser.AngleBracketContext):
        pass


    # Enter a parse tree produced by InputFormParser#Power.
    def enterPower(self, ctx:InputFormParser.PowerContext):
        pass

    # Exit a parse tree produced by InputFormParser#Power.
    def exitPower(self, ctx:InputFormParser.PowerContext):
        pass


    # Enter a parse tree produced by InputFormParser#ContextName.
    def enterContextName(self, ctx:InputFormParser.ContextNameContext):
        pass

    # Exit a parse tree produced by InputFormParser#ContextName.
    def exitContextName(self, ctx:InputFormParser.ContextNameContext):
        pass


    # Enter a parse tree produced by InputFormParser#SimpleContext.
    def enterSimpleContext(self, ctx:InputFormParser.SimpleContextContext):
        pass

    # Exit a parse tree produced by InputFormParser#SimpleContext.
    def exitSimpleContext(self, ctx:InputFormParser.SimpleContextContext):
        pass


    # Enter a parse tree produced by InputFormParser#CompoundContext.
    def enterCompoundContext(self, ctx:InputFormParser.CompoundContextContext):
        pass

    # Exit a parse tree produced by InputFormParser#CompoundContext.
    def exitCompoundContext(self, ctx:InputFormParser.CompoundContextContext):
        pass


    # Enter a parse tree produced by InputFormParser#NumberBaseN.
    def enterNumberBaseN(self, ctx:InputFormParser.NumberBaseNContext):
        pass

    # Exit a parse tree produced by InputFormParser#NumberBaseN.
    def exitNumberBaseN(self, ctx:InputFormParser.NumberBaseNContext):
        pass


    # Enter a parse tree produced by InputFormParser#NumberBaseTen.
    def enterNumberBaseTen(self, ctx:InputFormParser.NumberBaseTenContext):
        pass

    # Exit a parse tree produced by InputFormParser#NumberBaseTen.
    def exitNumberBaseTen(self, ctx:InputFormParser.NumberBaseTenContext):
        pass


    # Enter a parse tree produced by InputFormParser#numberLiteralPrecision.
    def enterNumberLiteralPrecision(self, ctx:InputFormParser.NumberLiteralPrecisionContext):
        pass

    # Exit a parse tree produced by InputFormParser#numberLiteralPrecision.
    def exitNumberLiteralPrecision(self, ctx:InputFormParser.NumberLiteralPrecisionContext):
        pass


    # Enter a parse tree produced by InputFormParser#numberLiteralExponent.
    def enterNumberLiteralExponent(self, ctx:InputFormParser.NumberLiteralExponentContext):
        pass

    # Exit a parse tree produced by InputFormParser#numberLiteralExponent.
    def exitNumberLiteralExponent(self, ctx:InputFormParser.NumberLiteralExponentContext):
        pass


    # Enter a parse tree produced by InputFormParser#OutNumbered.
    def enterOutNumbered(self, ctx:InputFormParser.OutNumberedContext):
        pass

    # Exit a parse tree produced by InputFormParser#OutNumbered.
    def exitOutNumbered(self, ctx:InputFormParser.OutNumberedContext):
        pass


    # Enter a parse tree produced by InputFormParser#OutUnnumbered.
    def enterOutUnnumbered(self, ctx:InputFormParser.OutUnnumberedContext):
        pass

    # Exit a parse tree produced by InputFormParser#OutUnnumbered.
    def exitOutUnnumbered(self, ctx:InputFormParser.OutUnnumberedContext):
        pass


    # Enter a parse tree produced by InputFormParser#SlotDigits.
    def enterSlotDigits(self, ctx:InputFormParser.SlotDigitsContext):
        pass

    # Exit a parse tree produced by InputFormParser#SlotDigits.
    def exitSlotDigits(self, ctx:InputFormParser.SlotDigitsContext):
        pass


    # Enter a parse tree produced by InputFormParser#SlotNamed.
    def enterSlotNamed(self, ctx:InputFormParser.SlotNamedContext):
        pass

    # Exit a parse tree produced by InputFormParser#SlotNamed.
    def exitSlotNamed(self, ctx:InputFormParser.SlotNamedContext):
        pass


    # Enter a parse tree produced by InputFormParser#SlotSequenceDigits.
    def enterSlotSequenceDigits(self, ctx:InputFormParser.SlotSequenceDigitsContext):
        pass

    # Exit a parse tree produced by InputFormParser#SlotSequenceDigits.
    def exitSlotSequenceDigits(self, ctx:InputFormParser.SlotSequenceDigitsContext):
        pass


    # Enter a parse tree produced by InputFormParser#SlotSequence.
    def enterSlotSequence(self, ctx:InputFormParser.SlotSequenceContext):
        pass

    # Exit a parse tree produced by InputFormParser#SlotSequence.
    def exitSlotSequence(self, ctx:InputFormParser.SlotSequenceContext):
        pass


    # Enter a parse tree produced by InputFormParser#SlotUnnamed.
    def enterSlotUnnamed(self, ctx:InputFormParser.SlotUnnamedContext):
        pass

    # Exit a parse tree produced by InputFormParser#SlotUnnamed.
    def exitSlotUnnamed(self, ctx:InputFormParser.SlotUnnamedContext):
        pass


    # Enter a parse tree produced by InputFormParser#ExpressionListed.
    def enterExpressionListed(self, ctx:InputFormParser.ExpressionListedContext):
        pass

    # Exit a parse tree produced by InputFormParser#ExpressionListed.
    def exitExpressionListed(self, ctx:InputFormParser.ExpressionListedContext):
        pass


    # Enter a parse tree produced by InputFormParser#AccessExpressionA.
    def enterAccessExpressionA(self, ctx:InputFormParser.AccessExpressionAContext):
        pass

    # Exit a parse tree produced by InputFormParser#AccessExpressionA.
    def exitAccessExpressionA(self, ctx:InputFormParser.AccessExpressionAContext):
        pass


    # Enter a parse tree produced by InputFormParser#AccessExpressionB.
    def enterAccessExpressionB(self, ctx:InputFormParser.AccessExpressionBContext):
        pass

    # Exit a parse tree produced by InputFormParser#AccessExpressionB.
    def exitAccessExpressionB(self, ctx:InputFormParser.AccessExpressionBContext):
        pass





"""
This module defines the mechanism by which errors are handled in FoxySheep.
"""

import sys

from antlr4.ParserRuleContext import ParserRuleContext
from tree import treeNode


def _print_err(*args, flush=False):
    sys.stderr.write(' '.join(map(str, args)) + '\n')
    if flush:
        sys.stderr.flush()


class ErrorBase:

    def __init__(self, parse_tree_node: ParserRuleContext=None,
                 ast_node: treeNode = None):
        self.parse_tree_node = parse_tree_node
        self.ast_node = ast_node

        class_name = type(self).__name__
        if class_name.endswith('Base'):
            class_name = class_name[:-4]
        elif class_name.endswith('_'):
            class_name = class_name[:-1]
        self.name = class_name

        self.message = 'Unknown error.'

    def get_line_number(self):
        """
        Returns the line number in the source text where the error occurs, or -1 if unknown.

        :return: line number, integer.
        """
        if self.parse_tree_node:
            start_token = self.parse_tree_node.start
            return start_token.getLine()
        # We don't have an associated parse_tree_node.
        return -1

    def get_char_position(self):
        """
        Returns the character position in the line in the source text where the error occurs, or -1 if unknown.

        :return: character position, integer.
        """
        if self.parse_tree_node:
            start_token = self.parse_tree_node.start
            return start_token.getCharPositionInLine()
        # We don't have an associated parse_tree_node.
        return -1

    def to_string(self):
        line = self.get_line_number()
        char = self.get_char_position()
        location = ''
        if line > -1 or char > -1:
            location = '{line}:{char}:'.format(line=line, char=char)

        return '{location}{error_name}: {message}'.format(
                location=location,
                error_name=self.name,
                message=self.message
                )


class UnknownFunctionError(ErrorBase):

    def __init__(self, identifier='', **kwargs):
        super().__init__(**kwargs)

        self.message = 'The function {identifier} is being called but is not ' \
                       'yet defined.'.format(identifier=identifier)


class NotImplementedError_(ErrorBase):

    def __init__(self, feature: str = '', **kwargs):
        super().__init__(**kwargs)

        self.message = '{feature} has not yet been implemented.'.format(
            feature=feature)


class ExternalExceptionError(ErrorBase):
    """
    An exception from an external library repackaged as an ErrorBase subclass.
    """

    def __init__(self, e: Exception, **kwargs):
        super().__init__(**kwargs)

        self._exception = e
        self.message = str(e)


def handle_error(error: ErrorBase):
    _print_err(error.to_string())
